# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from oslo_utils import units

from manila.share.drivers.dell_emc.plugins.unity import utils
from manila import test


class MockSP(object):
    def __init__(self, sp_id):
        self.sp_id = sp_id

    def get_id(self):
        return self.sp_id


SPA = MockSP('spa')
SPB = MockSP('spb')


class MockPort(object):
    def __init__(self, sp, port_id, mtu):
        self._sp = sp
        self.port_id = port_id
        self.mtu = mtu

    def get_id(self):
        return self.port_id

    @property
    def parent_storage_processor(self):
        return self._sp


SPA_ETH0 = MockPort(SPA, 'spa_eth0', 1500)
SPA_ETH1 = MockPort(SPA, 'spa_eth1', 9000)
SPB_ETH0 = MockPort(SPB, 'spb_eth0', 1500)
SPB_ETH1 = MockPort(SPB, 'spb_eth1', 9000)
SPA_LA1 = MockPort(SPA, 'spa_la_1', 1500)
SPB_LA1 = MockPort(SPB, 'spb_la_1', 1500)


@ddt.ddt
class TestUtils(test.TestCase):
    @ddt.data({'matcher': None,
               'matched': {'pool_1', 'pool_2', 'nas_server_pool'},
               'not_matched': set()},
              {'matcher': ['*'],
               'matched': {'pool_1', 'pool_2', 'nas_server_pool'},
               'not_matched': set()},
              {'matcher': ['pool_*'],
               'matched': {'pool_1', 'pool_2'},
               'not_matched': {'nas_server_pool'}},
              {'matcher': ['*pool'],
               'matched': {'nas_server_pool'},
               'not_matched': {'pool_1', 'pool_2'}},
              {'matcher': ['nas_server_pool'],
               'matched': {'nas_server_pool'},
               'not_matched': {'pool_1', 'pool_2'}},
              {'matcher': ['nas_*', 'pool_*'],
               'matched': {'pool_1', 'pool_2', 'nas_server_pool'},
               'not_matched': set()})
    def test_do_match(self, data):
        full = ['pool_1 ', ' pool_2', ' nas_server_pool ']
        matcher = data['matcher']
        expected_matched = data['matched']
        expected_not_matched = data['not_matched']

        matched, not_matched = utils.do_match(full, matcher)
        self.assertEqual(expected_matched, matched)
        self.assertEqual(expected_not_matched, not_matched)

    @ddt.data({'ports': [SPA_ETH0, SPB_ETH0],
               'ids_conf': None,
               'port_map': {'spa': {'spa_eth0'}, 'spb': {'spb_eth0'}},
               'unmanaged': set()},
              {'ports': [SPA_ETH0, SPB_ETH0],
               'ids_conf': ['   '],
               'port_map': {'spa': {'spa_eth0'}, 'spb': {'spb_eth0'}},
               'unmanaged': set()},
              {'ports': [SPA_ETH0, SPB_ETH0, SPA_ETH1],
               'ids_conf': ['spa*'],
               'port_map': {'spa': {'spa_eth0', 'spa_eth1'}},
               'unmanaged': {'spb_eth0'}},
              )
    @ddt.unpack
    def test_match_ports(self, ports, ids_conf, port_map, unmanaged):
        sp_ports_map, unmanaged_port_ids = utils.match_ports(ports,
                                                             ids_conf)
        self.assertEqual(port_map, sp_ports_map)
        self.assertEqual(unmanaged, unmanaged_port_ids)

    def test_find_ports_by_mtu(self):
        all_ports = [SPA_ETH0, SPB_ETH0, SPA_ETH1, SPB_ETH1, SPA_LA1,
                     SPB_LA1]
        port_ids_conf = '*'
        port_map = utils.find_ports_by_mtu(all_ports, port_ids_conf, 1500)
        self.assertEqual({'spa': {'spa_eth0', 'spa_la_1'},
                          'spb': {'spb_eth0', 'spb_la_1'}},
                         port_map)

    def test_gb_to_byte(self):
        self.assertEqual(3 * units.Gi, utils.gib_to_byte(3))

    def test_get_snapshot_id(self):
        snapshot = {'provider_location': '23047-ef2344-4563cvw-r4323cwed',
                    'id': 'test_id'}
        result = utils.get_snapshot_id(snapshot)
        expected = '23047-ef2344-4563cvw-r4323cwed'
        self.assertEqual(expected, result)

    def test_get_snapshot_id_without_pl(self):
        snapshot = {'provider_location': '', 'id': 'test_id'}
        result = utils.get_snapshot_id(snapshot)
        expected = 'test_id'
        self.assertEqual(expected, result)

    def test_get_nfs_share_id(self):
        nfs_share = {'export_locations':
                     [{'path': '10.10.1.12:/addf-97e-46c-8ac6-55922f',
                       'share_instance_id': 'e24-457e-47-12c6-gf345'}],
                     'share_proto': 'NFS', 'id': 'test_nfs_id'}
        result = utils.get_share_backend_id(nfs_share)
        expected = 'addf-97e-46c-8ac6-55922f'
        self.assertEqual(expected, result)

    def test_get_nfs_share_id_without_path(self):
        nfs_share = {'export_locations':
                     [{'path': '',
                       'share_instance_id': 'ev24-7e-4-12c6-g45245'}],
                     'share_proto': 'NFS', 'id': 'test_nfs_id'}
        result = utils.get_share_backend_id(nfs_share)
        expected = 'test_nfs_id'
        self.assertEqual(expected, result)

    def test_get_cifs_share_id(self):
        cifs_share = {'export_locations':
                      [{'path': '\\\\17.66.5.3\\bdf-h4e-42c-122c5-b212',
                        'share_instance_id': 'ev4-47e-48-126-gfbh452'}],
                      'share_proto': 'CIFS', 'id': 'test_cifs_id'}
        result = utils.get_share_backend_id(cifs_share)
        expected = 'bdf-h4e-42c-122c5-b212'
        self.assertEqual(expected, result)

    def test_get_cifs_share_id_without_path(self):
        cifs_share = {'export_locations':
                      [{'path': '',
                        'share_instance_id': 'ef4-47e-48-12c6-gf452'}],
                      'share_proto': 'CIFS', 'id': 'test_cifs_id'}
        result = utils.get_share_backend_id(cifs_share)
        expected = 'test_cifs_id'
        self.assertEqual(expected, result)



# Copyright (c) 2015 Quobyte Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Quobyte driver.

Manila shares are directly mapped to Quobyte volumes. The access to the
shares is provided by the Quobyte NFS proxy (a Ganesha NFS server).
"""

import math

from oslo_config import cfg
from oslo_log import log
from oslo_utils import units
import six

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers.quobyte import jsonrpc

LOG = log.getLogger(__name__)

quobyte_manila_share_opts = [
    cfg.StrOpt('quobyte_api_url',
               help='URL of the Quobyte API server (http or https)'),
    cfg.StrOpt('quobyte_api_ca',
               help='The X.509 CA file to verify the server cert.'),
    cfg.BoolOpt('quobyte_delete_shares',
                default=False,
                help='Actually deletes shares (vs. unexport)'),
    cfg.StrOpt('quobyte_api_username',
               default='admin',
               help='Username for Quobyte API server.'),
    cfg.StrOpt('quobyte_api_password',
               default='quobyte',
               secret=True,
               help='Password for Quobyte API server'),
    cfg.StrOpt('quobyte_volume_configuration',
               default='BASE',
               help='Name of volume configuration used for new shares.'),
    cfg.StrOpt('quobyte_default_volume_user',
               default='root',
               help='Default owning user for new volumes.'),
    cfg.StrOpt('quobyte_default_volume_group',
               default='root',
               help='Default owning group for new volumes.'),
    cfg.StrOpt('quobyte_export_path',
               default='/quobyte',
               help='Export path for shares of this bacckend. This needs '
                    'to match the quobyte-nfs services "Pseudo" option.'),
]

CONF = cfg.CONF
CONF.register_opts(quobyte_manila_share_opts)


class QuobyteShareDriver(driver.ExecuteMixin, driver.ShareDriver,):
    """Map share commands to Quobyte volumes.

    Version history:
        1.0     - Initial driver.
        1.0.1   - Adds ensure_share() implementation.
        1.1     - Adds extend_share() and shrink_share() implementation.
        1.2     - Adds update_access() implementation and related methods
        1.2.1   - Improved capacity calculation
        1.2.2   - Minor optimizations
        1.2.3   - Updated RPC layer for improved stability
        1.2.4   - Fixed handling updated QB API error codes
        1.2.5   - Fixed two quota handling bugs
        1.2.6   - Fixed volume resize and jsonrpc code style bugs
        1.2.7   - Add quobyte_export_path option
    """

    DRIVER_VERSION = '1.2.7'

    def __init__(self, *args, **kwargs):
        super(QuobyteShareDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(quobyte_manila_share_opts)
        self.backend_name = (self.configuration.safe_get('share_backend_name')
                             or CONF.share_backend_name or 'Quobyte')

    def _fetch_existing_access(self, context, share):
        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])
        result = self.rpc.call('getConfiguration', {})
        if result is None:
            raise exception.QBException(
                "Could not retrieve Quobyte configuration data!")
        tenant_configs = result['tenant_configuration']
        qb_access_list = []
        for tc in tenant_configs:
            for va in tc['volume_access']:
                if va['volume_uuid'] == volume_uuid:
                    a_level = constants.ACCESS_LEVEL_RW
                    if va['read_only']:
                        a_level = constants.ACCESS_LEVEL_RO
                    qb_access_list.append({
                        'access_to': va['restrict_to_network'],
                        'access_level': a_level,
                        'access_type': 'ip'
                    })
        return qb_access_list

    def do_setup(self, context):
        """Prepares the backend."""
        self.rpc = jsonrpc.JsonRpc(
            url=self.configuration.quobyte_api_url,
            ca_file=self.configuration.quobyte_api_ca,
            user_credentials=(
                self.configuration.quobyte_api_username,
                self.configuration.quobyte_api_password))

        try:
            self.rpc.call('getInformation', {})
        except Exception as exc:
            LOG.error("Could not connect to API: %s", exc)
            raise exception.QBException(
                _('Could not connect to API: %s') % exc)

    def _update_share_stats(self):
        total_gb, free_gb = self._get_capacities()

        data = dict(
            storage_protocol='NFS',
            vendor_name='Quobyte',
            share_backend_name=self.backend_name,
            driver_version=self.DRIVER_VERSION,
            total_capacity_gb=total_gb,
            free_capacity_gb=free_gb,
            reserved_percentage=self.configuration.reserved_share_percentage,
            reserved_snapshot_percentage=(
                self.configuration.reserved_share_from_snapshot_percentage
                or self.configuration.reserved_share_percentage))
        super(QuobyteShareDriver, self)._update_share_stats(data)

    def _get_capacities(self):
        result = self.rpc.call('getSystemStatistics', {})

        total = float(result['total_physical_capacity'])
        used = float(result['total_physical_usage'])
        LOG.info('Read capacity of %(cap)s bytes and '
                 'usage of %(use)s bytes from backend. ',
                 {'cap': total, 'use': used})
        free = total - used
        if free < 0:
            free = 0  # no space available
        free_replicated = free / self._get_qb_replication_factor()
        # floor numbers to nine digits (bytes)
        total = math.floor((total / units.Gi) * units.G) / units.G
        free = math.floor((free_replicated / units.Gi) * units.G) / units.G

        return total, free

    def _get_qb_replication_factor(self):
        result = self.rpc.call('getEffectiveVolumeConfiguration',
                               {'configuration_name': self.
                                configuration.quobyte_volume_configuration})
        return int(result['configuration']['volume_metadata_configuration']
                   ['replication_factor'])

    def check_for_setup_error(self):
        pass

    def get_network_allocations_number(self):
        return 0

    def _get_project_name(self, context, project_id):
        """Retrieve the project name.

        TODO (kaisers): retrieve the project name in order
        to store and use in the backend for better usability.
        """
        return project_id

    def _resize_share(self, share, new_size):
        newsize_bytes = new_size * units.Gi
        self.rpc.call('setQuota', {"quotas": [
            {"consumer":
                [{"type": "VOLUME",
                  "identifier": self._resolve_volume_name(share["name"],
                                                          share['project_id']),
                  "tenant_id": share["project_id"]}],
             "limits": [{"type": "LOGICAL_DISK_SPACE",
                        "value": newsize_bytes}]}
        ]})

    def _resolve_volume_name(self, volume_name, tenant_domain):
        """Resolve a volume name to the global volume uuid."""
        result = self.rpc.call('resolveVolumeName', dict(
            volume_name=volume_name,
            tenant_domain=tenant_domain), [jsonrpc.ERROR_ENOENT,
                                           jsonrpc.ERROR_ENTITY_NOT_FOUND])
        if result:
            return result['volume_uuid']
        return None  # not found

    def _subtract_access_lists(self, list_a, list_b):
        """Returns a list of elements in list_a that are not in list_b

        :param list_a: Base list of access rules
        :param list_b:  List of access rules not to be returned
        :return: List of elements of list_a not present in
        list_b
        """
        sub_tuples_list = [{"to": s.get('access_to'),
                            "type": s.get('access_type'),
                            "level": s.get('access_level')}
                           for s in list_b]
        return [r for r in list_a if (
            {"to": r.get("access_to"),
             "type": r.get("access_type"),
             "level": r.get("access_level")} not in sub_tuples_list)]

    def create_share(self, context, share, share_server=None):
        """Create or export a volume that is usable as a Manila share."""
        if share['share_proto'] != 'NFS':
            raise exception.QBException(
                _('Quobyte driver only supports NFS shares'))

        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])

        if not volume_uuid:
            # create tenant, expect ERROR_GARBAGE_ARGS if it already exists
            self.rpc.call('setTenant',
                          dict(tenant=dict(tenant_id=share['project_id'])),
                          expected_errors=[jsonrpc.ERROR_GARBAGE_ARGS])
            result = self.rpc.call('createVolume', dict(
                name=share['name'],
                tenant_domain=share['project_id'],
                root_user_id=self.configuration.quobyte_default_volume_user,
                root_group_id=self.configuration.quobyte_default_volume_group,
                configuration_name=(self.configuration.
                                    quobyte_volume_configuration)))
            volume_uuid = result['volume_uuid']

        result = self.rpc.call('exportVolume', dict(
            volume_uuid=volume_uuid,
            protocol='NFS'))

        self._resize_share(share, share['size'])

        return self._build_share_export_string(result)

    def delete_share(self, context, share, share_server=None):
        """Delete the corresponding Quobyte volume."""
        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])
        if not volume_uuid:
            LOG.warning("No volume found for "
                        "share %(project_id)s/%(name)s",
                        {"project_id": share['project_id'],
                         "name": share['name']})
            return

        if self.configuration.quobyte_delete_shares:
            self.rpc.call('deleteVolume', {'volume_uuid': volume_uuid})
        else:
            self.rpc.call('exportVolume', {"volume_uuid": volume_uuid,
                                           "remove_export": True,
                                           })

    def ensure_share(self, context, share, share_server=None):
        """Invoked to ensure that share is exported.

        :param context: The `context.RequestContext` object for the request
        :param share: Share instance that will be checked.
        :param share_server: Data structure with share server information.
        Not used by this driver.
        :returns: IP:<nfs_export_path> of share
        :raises:
            :ShareResourceNotFound: If the share instance cannot be found in
            the backend
        """

        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])

        LOG.debug("Ensuring Quobyte share %s", share['name'])

        if not volume_uuid:
            raise (exception.ShareResourceNotFound(
                share_id=share['id']))

        result = self.rpc.call('exportVolume', dict(
            volume_uuid=volume_uuid,
            protocol='NFS'))

        return self._build_share_export_string(result)

    def _allow_access(self, context, share, access, share_server=None):
        """Allow access to a share."""
        if access['access_type'] != 'ip':
            raise exception.InvalidShareAccess(
                _('Quobyte driver only supports ip access control'))

        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])
        ro = access['access_level'] == (constants.ACCESS_LEVEL_RO)
        call_params = {
            "volume_uuid": volume_uuid,
            "read_only": ro,
            "add_allow_ip": access['access_to']}
        self.rpc.call('exportVolume', call_params)

    def _build_share_export_string(self, rpc_result):
        return '%(nfs_server_ip)s:%(qb_exp_path)s%(nfs_export_path)s' % {
            "nfs_server_ip": rpc_result["nfs_server_ip"],
            "qb_exp_path": self.configuration.quobyte_export_path,
            "nfs_export_path": rpc_result["nfs_export_path"]}

    def _deny_access(self, context, share, access, share_server=None):
        """Remove white-list ip from a share."""
        if access['access_type'] != 'ip':
            LOG.debug('Quobyte driver only supports ip access control. '
                      'Ignoring deny access call for %s , %s',
                      share['name'],
                      self._get_project_name(context, share['project_id']))
            return

        volume_uuid = self._resolve_volume_name(share['name'],
                                                share['project_id'])
        call_params = {
            "volume_uuid": volume_uuid,
            "remove_allow_ip": access['access_to']}
        self.rpc.call('exportVolume', call_params)

    def extend_share(self, ext_share, ext_size, share_server=None):
        """Uses _resize_share to extend a share.

        :param ext_share: Share model.
        :param ext_size: New size of share (new_size > share['size']).
        :param share_server: Currently not used.
        """
        self._resize_share(share=ext_share, new_size=ext_size)

    def shrink_share(self, shrink_share, shrink_size, share_server=None):
        """Uses _resize_share to shrink a share.

        Quobyte uses soft quotas. If a shares current size is bigger than
        the new shrunken size no data is lost. Data can be continuously read
        from the share but new writes receive out of disk space replies.

        :param shrink_share: Share model.
        :param shrink_size: New size of share (new_size < share['size']).
        :param share_server: Currently not used.
        """
        self._resize_share(share=shrink_share, new_size=shrink_size)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share.

        Two different cases are supported in here:
        1. Recovery after error - 'access_rules' contains all access_rules,
        'add_rules' and 'delete_rules' are empty. Driver should apply all
        access rules for given share.

        2. Adding/Deleting of several access rules - 'access_rules' contains
        all access_rules, 'add_rules' and 'delete_rules' contain rules which
        should be added/deleted. Driver can ignore rules in 'access_rules' and
        apply only rules from 'add_rules' and 'delete_rules'.

        :param context: Current context
        :param share: Share model with share data.
        :param access_rules: All access rules for given share
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: None or Share server model
        :raises If all of the *_rules params are None the method raises an
        InvalidShareAccess exception
        """
        if (add_rules or delete_rules):
            # Handling access rule update
            for d_rule in delete_rules:
                self._deny_access(context, share, d_rule)
            for a_rule in add_rules:
                self._allow_access(context, share, a_rule)
        else:
            if not access_rules:
                LOG.warning("No access rules provided in update_access.")
            else:
                # Handling access rule recovery
                existing_rules = self._fetch_existing_access(context, share)

                missing_rules = self._subtract_access_lists(access_rules,
                                                            existing_rules)
                for a_rule in missing_rules:
                    LOG.debug("Adding rule %s in recovery.",
                              six.text_type(a_rule))
                    self._allow_access(context, share, a_rule)

                superfluous_rules = self._subtract_access_lists(existing_rules,
                                                                access_rules)
                for d_rule in superfluous_rules:
                    LOG.debug("Removing rule %s in recovery.",
                              six.text_type(d_rule))
                    self._deny_access(context, share, d_rule)



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from manila.api.views import scheduler_stats
from manila import test


POOL1 = {
    'name': 'host1@backend1#pool1',
    'host': 'host1',
    'backend': 'backend1',
    'pool': 'pool1',
    'other': 'junk',
    'capabilities': {
        'pool_name': 'pool1',
        'driver_handles_share_servers': False,
        'qos': 'False',
        'timestamp': '2015-03-15T19:15:42.611690',
        'allocated_capacity_gb': 5,
        'total_capacity_gb': 10,
    },
}
POOL2 = {
    'name': 'host1@backend1#pool2',
    'host': 'host1',
    'backend': 'backend1',
    'pool': 'pool2',
    'capabilities': {
        'pool_name': 'pool2',
        'driver_handles_share_servers': False,
        'qos': 'False',
        'timestamp': '2015-03-15T19:15:42.611690',
        'allocated_capacity_gb': 15,
        'total_capacity_gb': 20,
    },
}
POOLS = [POOL1, POOL2]

POOLS_DETAIL_VIEW = {
    'pools': [
        {
            'name': 'host1@backend1#pool1',
            'host': 'host1',
            'backend': 'backend1',
            'pool': 'pool1',
            'capabilities': {
                'pool_name': 'pool1',
                'driver_handles_share_servers': False,
                'qos': 'False',
                'timestamp': '2015-03-15T19:15:42.611690',
                'allocated_capacity_gb': 5,
                'total_capacity_gb': 10,
            },
        }, {
            'name': 'host1@backend1#pool2',
            'host': 'host1',
            'backend': 'backend1',
            'pool': 'pool2',
            'capabilities': {
                'pool_name': 'pool2',
                'driver_handles_share_servers': False,
                'qos': 'False',
                'timestamp': '2015-03-15T19:15:42.611690',
                'allocated_capacity_gb': 15,
                'total_capacity_gb': 20,
            }
        }
    ]
}


class ViewBuilderTestCase(test.TestCase):

    def setUp(self):
        super(ViewBuilderTestCase, self).setUp()
        self.builder = scheduler_stats.ViewBuilder()

    def test_pools(self):

        result = self.builder.pools(POOLS)

        # Remove capabilities for summary view
        expected = copy.deepcopy(POOLS_DETAIL_VIEW)
        for pool in expected['pools']:
            del pool['capabilities']

        self.assertDictEqual(expected, result)

    def test_pools_with_details(self):

        result = self.builder.pools(POOLS, detail=True)

        expected = POOLS_DETAIL_VIEW
        self.assertDictEqual(expected, result)



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os import path
from unittest import mock
import yaml

from oslo_log import log

LOG = log.getLogger(__name__)

patch_system = mock.patch('storops.UnitySystem')


def load_yaml(file_name):
    yaml_file = '{}/{}'.format(path.dirname(path.abspath(__file__)), file_name)
    with open(yaml_file) as f:
        res = yaml.safe_load(f)
        LOG.debug('Loaded yaml mock objects from %s.', yaml_file)
    return res


patch_find_ports_by_mtu = mock.patch('manila.share.drivers.dell_emc.plugins.'
                                     'unity.utils.find_ports_by_mtu')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fix 'project_share_type_quotas' unique constraint

Revision ID: 829a09b0ddd4
Revises: b516de97bfee
Create Date: 2017-10-12 20:15:51.267488

"""

# revision identifiers, used by Alembic.
revision = '829a09b0ddd4'
down_revision = 'b516de97bfee'

from alembic import op

TABLE_NAME = 'project_share_type_quotas'
UNIQUE_CONSTRAINT_NAME = 'uc_quotas_per_share_types'
ST_FK_NAME = 'share_type_id_fk'


def upgrade():
    op.drop_constraint(ST_FK_NAME, TABLE_NAME, type_='foreignkey')
    op.drop_constraint(UNIQUE_CONSTRAINT_NAME, TABLE_NAME, type_='unique')
    op.create_foreign_key(
        ST_FK_NAME, TABLE_NAME, 'share_types', ['share_type_id'], ['id'])
    op.create_unique_constraint(
        UNIQUE_CONSTRAINT_NAME, TABLE_NAME,
        ['share_type_id', 'resource', 'deleted', 'project_id'])


def downgrade():
    # NOTE(vponomaryov): no need to implement old behaviour as it was bug, and,
    # moreover, not compatible with data from upgraded version.
    pass



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from manila.common import constants as const
from manila import exception
from manila.i18n import _

LOG = log.getLogger(__name__)


class DockerCIFSHelper(object):
    def __init__(self, container_helper, *args, **kwargs):
        super(DockerCIFSHelper, self).__init__()
        self.share = kwargs.get("share")
        self.conf = kwargs.get("config")
        self.container = container_helper

    def create_share(self, server_id):
        share_name = self.share.share_id
        cmd = ["net", "conf", "addshare", share_name,
               "/shares/%s" % share_name, "writeable=y"]
        if self.conf.container_cifs_guest_ok:
            cmd.append("guest_ok=y")
        else:
            cmd.append("guest_ok=n")
        self.container.execute(server_id, cmd)
        parameters = {
            "browseable": "yes",
            "create mask": "0755",
            "read only": "no",
        }
        for param, value in parameters.items():
            self.container.execute(
                server_id,
                ["net", "conf", "setparm", share_name, param, value]
            )
        # TODO(tbarron): pass configured address family when we support IPv6
        address = self.container.fetch_container_address(
            server_id, address_family='inet')
        return r"//%(addr)s/%(name)s" % {"addr": address, "name": share_name}

    def delete_share(self, server_id, share_name, ignore_errors=False):
        self.container.execute(
            server_id,
            ["net", "conf", "delshare", share_name],
            ignore_errors=ignore_errors
        )

    def _get_access_group(self, access_level):
        if access_level == const.ACCESS_LEVEL_RO:
            access = "read list"
        elif access_level == const.ACCESS_LEVEL_RW:
            access = "valid users"
        else:
            raise exception.InvalidShareAccessLevel(level=access_level)
        return access

    def _get_existing_users(self, server_id, share_name, access):
        result = self.container.execute(
            server_id,
            ["net", "conf", "getparm", share_name, access],
            ignore_errors=True
        )
        if result:
            return result[0].rstrip('\n')
        else:
            return ""

    def _set_users(self, server_id, share_name, access, users_to_set):
        self.container.execute(
            server_id,
            ["net", "conf", "setparm", share_name, access, users_to_set]
        )

    def _allow_access(self, share_name, server_id, user_to_allow,
                      access_level):
        access = self._get_access_group(access_level)
        try:
            existing_users = self._get_existing_users(server_id, share_name,
                                                      access)
        except TypeError:
            users_to_allow = user_to_allow
        else:
            users_to_allow = " ".join([existing_users, user_to_allow])
        self._set_users(server_id, share_name, access, users_to_allow)

    def _deny_access(self, share_name, server_id, user_to_deny,
                     access_level):
        access = self._get_access_group(access_level)
        try:
            existing_users = self._get_existing_users(server_id, share_name,
                                                      access)
        except TypeError:
            LOG.warning("Can't access smbd at share %s.", share_name)
            return
        else:
            allowed_users = " ".join(sorted(set(existing_users.split()) -
                                     set([user_to_deny])))
            if allowed_users != existing_users:
                self._set_users(server_id, share_name, access, allowed_users)

    def update_access(self, server_id, share_name, access_rules,
                      add_rules=None, delete_rules=None):

        def _rule_updater(rules, action, override_type_check=False):
            for rule in rules:
                access_level = rule['access_level']
                access_type = rule['access_type']
                # (aovchinnikov): override_type_check is used to ensure
                # broken rules deletion.
                if access_type == 'user' or override_type_check:
                    action(share_name, server_id, rule['access_to'],
                           access_level)
                else:
                    msg = _("Access type '%s' is not supported by the "
                            "driver.") % access_type
                    raise exception.InvalidShareAccess(reason=msg)

        if not (add_rules or delete_rules):
            # clean all users first.
            self.container.execute(
                server_id,
                ["net", "conf", "setparm", share_name, "valid users", ""]
            )
            _rule_updater(access_rules or [], self._allow_access)
            return
        _rule_updater(add_rules or [], self._allow_access)
        _rule_updater(delete_rules or [], self._deny_access,
                      override_type_check=True)



# Copyright 2010-2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from manila.api import common
from manila import utils


class ViewBuilder(common.ViewBuilder):
    """OpenStack API base limits view builder."""

    _collection_name = "limits"
    _detail_version_modifiers = [
        "add_share_replica_quotas",
        "add_share_group_quotas"
    ]

    def build(self, request, rate_limits, absolute_limits):
        rate_limits = self._build_rate_limits(rate_limits)
        absolute_limits = self._build_absolute_limits(request, absolute_limits)

        output = {
            "limits": {
                "rate": rate_limits,
                "absolute": absolute_limits,
            },
        }

        return output

    def _build_absolute_limits(self, request, absolute_limits):
        """Builder for absolute limits.

        absolute_limits should be given as a dict of limits.
        For example: {"limit": {"shares": 10, "gigabytes": 1024},
                      "in_use": {"shares": 8, "gigabytes": 256}}.
        """
        limit_names = {
            "limit": {
                "gigabytes": ["maxTotalShareGigabytes"],
                "snapshot_gigabytes": ["maxTotalSnapshotGigabytes"],
                "shares": ["maxTotalShares"],
                "snapshots": ["maxTotalShareSnapshots"],
                "share_networks": ["maxTotalShareNetworks"],
            },
            "in_use": {
                "shares": ["totalSharesUsed"],
                "snapshots": ["totalShareSnapshotsUsed"],
                "share_networks": ["totalShareNetworksUsed"],
                "gigabytes": ["totalShareGigabytesUsed"],
                "snapshot_gigabytes": ["totalSnapshotGigabytesUsed"],
            },
        }
        limits = {}
        self.update_versioned_resource_dict(request, limit_names,
                                            absolute_limits)
        for mapping_key in limit_names.keys():
            for k, v in absolute_limits.get(mapping_key, {}).items():
                if k in limit_names.get(mapping_key, []) and v is not None:
                    for name in limit_names[mapping_key][k]:
                        limits[name] = v
        return limits

    def _build_rate_limits(self, rate_limits):
        limits = []
        for rate_limit in rate_limits:
            _rate_limit_key = None
            _rate_limit = self._build_rate_limit(rate_limit)

            # check for existing key
            for limit in limits:
                if (limit["uri"] == rate_limit["URI"] and
                        limit["regex"] == rate_limit["regex"]):
                    _rate_limit_key = limit
                    break

            # ensure we have a key if we didn't find one
            if not _rate_limit_key:
                _rate_limit_key = {
                    "uri": rate_limit["URI"],
                    "regex": rate_limit["regex"],
                    "limit": [],
                }
                limits.append(_rate_limit_key)

            _rate_limit_key["limit"].append(_rate_limit)

        return limits

    def _build_rate_limit(self, rate_limit):
        _get_utc = datetime.datetime.utcfromtimestamp
        next_avail = _get_utc(rate_limit["resetTime"])
        return {
            "verb": rate_limit["verb"],
            "value": rate_limit["value"],
            "remaining": int(rate_limit["remaining"]),
            "unit": rate_limit["unit"],
            "next-available": utils.isotime(at=next_avail),
        }

    @common.ViewBuilder.versioned_method("2.58")
    def add_share_group_quotas(self, request, limit_names, absolute_limits):
        limit_names["limit"]["share_groups"] = ["maxTotalShareGroups"]
        limit_names["limit"]["share_group_snapshots"] = (
            ["maxTotalShareGroupSnapshots"])
        limit_names["in_use"]["share_groups"] = ["totalShareGroupsUsed"]
        limit_names["in_use"]["share_group_snapshots"] = (
            ["totalShareGroupSnapshotsUsed"])

    @common.ViewBuilder.versioned_method("2.53")
    def add_share_replica_quotas(self, request, limit_names, absolute_limits):
        limit_names["limit"]["share_replicas"] = ["maxTotalShareReplicas"]
        limit_names["limit"]["replica_gigabytes"] = (
            ["maxTotalReplicaGigabytes"])
        limit_names["in_use"]["share_replicas"] = ["totalShareReplicasUsed"]
        limit_names["in_use"]["replica_gigabytes"] = (
            ["totalReplicaGigabytesUsed"])



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from manila.network.linux import interface
from manila.network.linux import ip_lib
from manila import test
from manila.tests import conf_fixture
from manila.tests import fake_network
from manila import utils


class BaseChild(interface.LinuxInterfaceDriver):
    def plug(self, *args):
        pass

    def unplug(self, *args):
        pass


FakeSubnet = {
    'cidr': '192.168.1.1/24',
}


FakeAllocation = {
    'subnet': FakeSubnet,
    'ip_address': '192.168.1.2',
    'ip_version': 4,
}


FakePort = {
    'id': 'abcdef01-1234-5678-90ab-ba0987654321',
    'fixed_ips': [FakeAllocation],
    'device_id': 'cccccccc-cccc-cccc-cccc-cccccccccccc',
}


class TestBase(test.TestCase):
    def setUp(self):
        super(TestBase, self).setUp()
        self.conf = conf_fixture.CONF
        self.conf.register_opts(interface.OPTS)
        self.ip_dev_p = mock.patch.object(ip_lib, 'IPDevice')
        self.ip_dev = self.ip_dev_p.start()
        self.ip_p = mock.patch.object(ip_lib, 'IPWrapper')
        self.ip = self.ip_p.start()
        self.device_exists_p = mock.patch.object(ip_lib, 'device_exists')
        self.device_exists = self.device_exists_p.start()
        self.addCleanup(self.ip_dev_p.stop)
        self.addCleanup(self.ip_p.stop)
        self.addCleanup(self.device_exists_p.stop)


class TestABCDriver(TestBase):

    def test_verify_abs_class_has_abs_methods(self):

        class ICanNotBeInstancetiated(interface.LinuxInterfaceDriver):
            pass

        try:
            # pylint: disable=abstract-class-instantiated
            ICanNotBeInstancetiated()
        except TypeError:
            pass
        except Exception as e:
            self.fail("Unexpected exception thrown: '%s'" % e)
        else:
            self.fail("ExpectedException 'TypeError' not thrown.")

    def test_get_device_name(self):
        bc = BaseChild()
        device_name = bc.get_device_name(FakePort)
        self.assertEqual('tapabcdef01-12', device_name)

    def test_l3_init(self):
        addresses = [dict(ip_version=4, scope='global',
                          dynamic=False, cidr='172.16.77.240/24')]
        self.ip_dev().addr.list = mock.Mock(return_value=addresses)

        bc = BaseChild()
        self.mock_object(bc, '_remove_outdated_interfaces')

        ns = '12345678-1234-5678-90ab-ba0987654321'
        bc.init_l3('tap0', ['192.168.1.2/24'], namespace=ns,
                   clear_cidrs=['192.168.0.0/16'])
        self.ip_dev.assert_has_calls(
            [mock.call('tap0', namespace=ns),
             mock.call().route.clear_outdated_routes('192.168.0.0/16'),
             mock.call().addr.list(scope='global', filters=['permanent']),
             mock.call().addr.add(4, '192.168.1.2/24', '192.168.1.255'),
             mock.call().addr.delete(4, '172.16.77.240/24'),
             mock.call().route.pullup_route('tap0')])
        bc._remove_outdated_interfaces.assert_called_with(self.ip_dev())

    def test__remove_outdated_interfaces(self):
        device = fake_network.FakeDevice(
            'foobarquuz', [dict(ip_version=4, cidr='1.0.0.0/27')])
        devices = [fake_network.FakeDevice('foobar')]
        self.ip().get_devices = mock.Mock(return_value=devices)

        bc = BaseChild()
        self.mock_object(bc, 'unplug')

        bc._remove_outdated_interfaces(device)
        bc.unplug.assert_called_once_with('foobar')

    def test__get_set_of_device_cidrs(self):
        device = fake_network.FakeDevice('foo')
        expected = set(('1.0.0.0/27', '2.0.0.0/27'))

        bc = BaseChild()
        result = bc._get_set_of_device_cidrs(device)

        self.assertEqual(expected, result)

    def test__get_set_of_device_cidrs_exception(self):
        device = fake_network.FakeDevice('foo')
        self.mock_object(device.addr, 'list', mock.Mock(
            side_effect=Exception('foo does not exist')))

        bc = BaseChild()
        result = bc._get_set_of_device_cidrs(device)

        self.assertEqual(set(), result)


class TestNoopInterfaceDriver(TestBase):

    def test_init_l3(self):
        self.ip.assert_not_called()
        self.ip_dev.assert_not_called()

    def test_plug(self):
        self.ip.assert_not_called()
        self.ip_dev.assert_not_called()

    def test_unplug(self):
        self.ip.assert_not_called()
        self.ip_dev.assert_not_called()


class TestOVSInterfaceDriver(TestBase):

    def test_get_device_name(self):
        br = interface.OVSInterfaceDriver()
        device_name = br.get_device_name(FakePort)
        self.assertEqual('tapabcdef01-12', device_name)

    def test_plug_no_ns(self):
        self._test_plug()

    def test_plug_with_ns(self):
        self._test_plug(namespace='01234567-1234-1234-99')

    def test_plug_alt_bridge(self):
        self._test_plug(bridge='br-foo')

    def _test_plug(self, additional_expectation=None, bridge=None,
                   namespace=None):
        if additional_expectation is None:
            additional_expectation = []
        if not bridge:
            bridge = 'br-int'

        def device_exists(dev, namespace=None):
            return dev == bridge

        vsctl_cmd = ['ovs-vsctl', '--', '--may-exist', 'add-port',
                     bridge, 'tap0', '--', 'set', 'Interface', 'tap0',
                     'type=internal', '--', 'set', 'Interface', 'tap0',
                     'external-ids:iface-id=port-1234', '--', 'set',
                     'Interface', 'tap0',
                     'external-ids:iface-status=active', '--', 'set',
                     'Interface', 'tap0',
                     'external-ids:attached-mac=aa:bb:cc:dd:ee:ff']

        with mock.patch.object(utils, 'execute') as execute:
            ovs = interface.OVSInterfaceDriver()
            self.device_exists.side_effect = device_exists
            ovs.plug('tap0',
                     'port-1234',
                     'aa:bb:cc:dd:ee:ff',
                     bridge=bridge,
                     namespace=namespace)
            execute.assert_called_once_with(*vsctl_cmd, run_as_root=True)

        expected = [mock.call(),
                    mock.call().device('tap0'),
                    mock.call().device().link.set_address('aa:bb:cc:dd:ee:ff')]
        expected.extend(additional_expectation)
        if namespace:
            expected.extend(
                [mock.call().ensure_namespace(namespace),
                 mock.call().ensure_namespace().add_device_to_namespace(
                     mock.ANY)])
        expected.extend([mock.call().device().link.set_up()])

        self.ip.assert_has_calls(expected)

    def test_plug_reset_mac(self):
        fake_mac_addr = 'aa:bb:cc:dd:ee:ff'
        self.device_exists.return_value = True

        self.ip().device().link.address = mock.Mock(return_value=fake_mac_addr)
        ovs = interface.OVSInterfaceDriver()
        ovs.plug('tap0',
                 'port-1234',
                 'ff:ee:dd:cc:bb:aa',
                 bridge='br-int')
        expected = [mock.call(),
                    mock.call().device('tap0'),
                    mock.call().device().link.set_address('ff:ee:dd:cc:bb:aa'),
                    mock.call().device().link.set_up()]
        self.ip.assert_has_calls(expected)

    def test_unplug(self, bridge=None):
        if not bridge:
            bridge = 'br-int'
        with mock.patch('manila.network.linux.ovs_lib.OVSBridge') as ovs_br:
            ovs = interface.OVSInterfaceDriver()
            ovs.unplug('tap0')
            ovs_br.assert_has_calls([mock.call(bridge),
                                     mock.call().delete_port('tap0')])


class TestBridgeInterfaceDriver(TestBase):
    def test_get_device_name(self):
        br = interface.BridgeInterfaceDriver()
        device_name = br.get_device_name(FakePort)
        self.assertEqual('ns-abcdef01-12', device_name)

    def test_plug_no_ns(self):
        self._test_plug()

    def test_plug_with_ns(self):
        self._test_plug(namespace='01234567-1234-1234-99')

    def _test_plug(self, namespace=None, mtu=None):
        def device_exists(device, root_helper=None, namespace=None):
            return device.startswith('brq')

        root_veth = mock.Mock()
        ns_veth = mock.Mock()

        self.ip().add_veth = mock.Mock(return_value=(root_veth, ns_veth))

        self.device_exists.side_effect = device_exists
        br = interface.BridgeInterfaceDriver()
        mac_address = 'aa:bb:cc:dd:ee:ff'
        br.plug('ns-0',
                'port-1234',
                mac_address,
                namespace=namespace)

        ip_calls = [mock.call(),
                    mock.call().add_veth('tap0', 'ns-0', namespace2=namespace)]
        ns_veth.assert_has_calls([mock.call.link.set_address(mac_address)])

        self.ip.assert_has_calls(ip_calls)

        root_veth.assert_has_calls([mock.call.link.set_up()])
        ns_veth.assert_has_calls([mock.call.link.set_up()])

    def test_plug_dev_exists(self):
        self.device_exists.return_value = True
        with mock.patch('manila.network.linux.interface.LOG.warning') as log:
            br = interface.BridgeInterfaceDriver()
            br.plug('port-1234',
                    'tap0',
                    'aa:bb:cc:dd:ee:ff')
            self.ip_dev.assert_has_calls([])
            self.assertEqual(1, log.call_count)

    def test_unplug_no_device(self):
        self.device_exists.return_value = False
        self.ip_dev().link.delete.side_effect = RuntimeError
        with mock.patch('manila.network.linux.interface.LOG') as log:
            br = interface.BridgeInterfaceDriver()
            br.unplug('tap0')
            [mock.call(), mock.call('tap0'), mock.call().link.delete()]
            self.assertEqual(1, log.error.call_count)

    def test_unplug(self):
        self.device_exists.return_value = True
        with mock.patch('manila.network.linux.interface.LOG.debug') as log:
            br = interface.BridgeInterfaceDriver()
            br.unplug('tap0')
            self.assertTrue(log.called)
        self.ip_dev.assert_has_calls([mock.call('tap0', None),
                                      mock.call().link.delete()])



# Copyright 2015 Alex Meade
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
from unittest import mock

import ddt
from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import webob

from manila.api.openstack import wsgi
import manila.api.v2.share_groups as share_groups
from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila.share import share_types
from manila.share_group import api as share_group_api
from manila.share_group import share_group_types
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


CONF = cfg.CONF
SG_GRADUATION_VERSION = '2.55'


@ddt.ddt
class ShareGroupAPITest(test.TestCase):
    """Consistency Groups API Test suite."""

    def setUp(self):
        super(ShareGroupAPITest, self).setUp()
        self.controller = share_groups.ShareGroupController()
        self.resource_name = self.controller.resource_name
        self.fake_share_type = {'id': uuidutils.generate_uuid()}
        self.fake_share_group_type = {
            'id': uuidutils.generate_uuid()}
        self.api_version = '2.34'
        self.request = fakes.HTTPRequest.blank(
            '/share-groups', version=self.api_version, experimental=True)
        self.flags(transport_url='rabbit://fake:fake@mqhost:5672')
        self.admin_context = context.RequestContext('admin', 'fake', True)
        self.member_context = context.RequestContext('fake', 'fake')
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.context = self.request.environ['manila.context']
        self.mock_object(share_group_types, 'get_default',
                         mock.Mock(return_value=self.fake_share_group_type))
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))

    def _get_context(self, role):
        return getattr(self, '%s_context' % role)

    def _setup_share_group_data(self, share_group=None, version='2.31'):
        if share_group is None:
            share_group = db_utils.create_share_group(
                status=constants.STATUS_AVAILABLE)
        path = '/v2/fake/share-groups/%s/action' % share_group['id']
        req = fakes.HTTPRequest.blank(path, script_name=path, version=version)
        req.headers[wsgi.API_VERSION_REQUEST_HEADER] = version
        req.headers[wsgi.EXPERIMENTAL_API_REQUEST_HEADER] = 'True'

        return share_group, req

    def _get_fake_share_group(self, ctxt=None, **values):
        if ctxt is None:
            ctxt = self.context

        share_group_db_dict = {
            'id': 'fake_id',
            'user_id': 'fakeuser',
            'project_id': 'fakeproject',
            'status': constants.STATUS_CREATING,
            'name': 'fake name',
            'description': 'fake description',
            'host': None,
            'availability_zone': None,
            'consistent_snapshot_support': None,
            'source_share_group_snapshot_id': None,
            'share_group_type_id': self.fake_share_group_type.get('id'),
            'share_network_id': uuidutils.generate_uuid(),
            'share_server_id': uuidutils.generate_uuid(),
            'share_types': [],
            'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
        }

        share_group_db_dict.update(**values)

        expected_share_group = {
            'id': share_group_db_dict['id'],
            'project_id': share_group_db_dict['project_id'],
            'status': share_group_db_dict['status'],
            'name': share_group_db_dict['name'],
            'description': share_group_db_dict['description'],
            'host': share_group_db_dict['host'],
            'availability_zone': share_group_db_dict['availability_zone'],
            'consistent_snapshot_support': share_group_db_dict[
                'consistent_snapshot_support'],
            'source_share_group_snapshot_id': share_group_db_dict[
                'source_share_group_snapshot_id'],
            'share_group_type_id': share_group_db_dict['share_group_type_id'],
            'share_network_id': share_group_db_dict['share_network_id'],
            'share_server_id': share_group_db_dict['share_server_id'],
            'share_types': [st['share_type_id']
                            for st in share_group_db_dict.get('share_types')],
            'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
            'links': mock.ANY,
        }
        if not ctxt.is_admin:
            del expected_share_group['share_server_id']

        return share_group_db_dict, expected_share_group

    def _get_fake_simple_share_group(self, **values):
        share_group = {'id': 'fake_id', 'name': None}
        share_group.update(**values)
        expected_share_group = copy.deepcopy(share_group)
        expected_share_group['links'] = mock.ANY
        return share_group, expected_share_group

    def _get_fake_custom_request_and_context(self, microversion, experimental):
        req = fakes.HTTPRequest.blank(
            '/share-groups', version=microversion, experimental=experimental)
        req_context = req.environ['manila.context']
        return req, req_context

    @ddt.data({'microversion': '2.34', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_create(self, microversion, experimental):
        fake, expected = self._get_fake_share_group()
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake))
        req, req_context = self._get_fake_custom_request_and_context(
            microversion, experimental)
        body = {"share_group": {}}

        res_dict = self.controller.create(req, body)

        self.controller.share_group_api.create.assert_called_once_with(
            req_context, share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.assertEqual(expected, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'create')

    def test_group_create_invalid_group_snapshot_state(self):
        fake_snap_id = uuidutils.generate_uuid()
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(side_effect=exception.InvalidShareGroupSnapshot(
                reason='bad status',
            )))
        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id
            }
        }

        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller.create, self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_no_default_share_type(self):
        fake_group, expected_group = self._get_fake_share_group()
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=None))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_group))
        body = {"share_group": {}}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.create, self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_no_default_group_type(self):
        fake_group, expected_group = self._get_fake_share_group()
        self.mock_object(
            share_group_types, 'get_default', mock.Mock(return_value=None))
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(return_value=fake_group))
        body = {"share_group": {}}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.create, self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_group_type_specified(self):
        fake_share_group, expected_group = self._get_fake_share_group()
        self.mock_object(
            share_group_types, 'get_default', mock.Mock(return_value=None))
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(return_value=fake_share_group))
        body = {
            "share_group": {
                "share_group_type_id": self.fake_share_group_type.get('id'),
            }
        }

        self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_invalid_group_type_specified(self):
        fake_share_group, expected_share_group = self._get_fake_share_group()
        self.mock_object(
            share_group_types, 'get_default', mock.Mock(return_value=None))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_share_group))
        body = {"share_group": {"group_type_id": "invalid"}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_az(self):
        fake_az_name = 'fake_az_name'
        fake_az_id = 'fake_az_id'
        fake_share_group, expected_share_group = self._get_fake_share_group(
            availability_zone_id=fake_az_id)
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(return_value=fake_share_group))
        self.mock_object(
            share_groups.db, 'availability_zone_get',
            mock.Mock(return_value=type(
                'FakeAZ', (object, ), {
                    'id': fake_az_id,
                    'name': fake_az_name,
                })))

        body = {"share_group": {"availability_zone": fake_az_name}}

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, availability_zone_id=fake_az_id,
            availability_zone=fake_az_name,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        share_groups.db.availability_zone_get.assert_called_once_with(
            self.context, fake_az_name)
        self.assertEqual(expected_share_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_az_and_source_share_group_snapshot(self):
        fake_az_name = 'fake_az_name'
        fake_az_id = 'fake_az_id'
        fake_share_group, expected_share_group = self._get_fake_share_group(
            availability_zone_id=fake_az_id)
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(return_value=fake_share_group))
        self.mock_object(
            share_groups.db, 'availability_zone_get',
            mock.Mock(return_value=type(
                'FakeAZ', (object, ), {
                    'id': fake_az_id,
                    'name': fake_az_name,
                })))

        body = {"share_group": {
            "availability_zone": fake_az_name,
            "source_share_group_snapshot_id": 'fake_sgs_id',
        }}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.create,
            self.request, body)

        self.controller.share_group_api.create.assert_not_called()
        share_groups.db.availability_zone_get.assert_not_called()
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_nonexistent_az(self):
        fake_az_name = 'fake_az_name'
        fake_az_id = 'fake_az_id'
        fake_share_group, expected_share_group = self._get_fake_share_group(
            availability_zone_id=fake_az_id)
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(return_value=fake_share_group))
        self.mock_object(
            share_groups.db, 'availability_zone_get',
            mock.Mock(
                side_effect=exception.AvailabilityZoneNotFound(id=fake_az_id)))

        body = {"share_group": {"availability_zone": fake_az_name}}

        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.create, self.request, body)

        self.assertEqual(0, self.controller.share_group_api.create.call_count)
        share_groups.db.availability_zone_get.assert_called_once_with(
            self.context, fake_az_name)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_name(self):
        fake_name = 'fake_name'
        fake_share_group, expected_share_group = self._get_fake_share_group(
            name=fake_name)
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_share_group))
        body = {"share_group": {"name": fake_name}}

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, name=fake_name,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.assertEqual(expected_share_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_description(self):
        fake_description = 'fake_description'
        fake_share_group, expected_share_group = self._get_fake_share_group(
            description=fake_description)
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_share_group))
        body = {"share_group": {"description": fake_description}}

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, description=fake_description,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.assertEqual(expected_share_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_share_types(self):
        fake_share_types = [{"share_type_id": self.fake_share_type['id']}]
        fake_group, expected_group = self._get_fake_share_group(
            share_types=fake_share_types)
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_group))
        body = {
            "share_group": {
                "share_types": [self.fake_share_type['id']]
            }
        }

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.assertEqual(expected_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_sg_create_with_source_sg_snapshot_id_and_share_network(self):
        fake_snap_id = uuidutils.generate_uuid()
        fake_net_id = uuidutils.generate_uuid()
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        mock_api_call = self.mock_object(
            self.controller.share_group_api, 'create')
        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id,
                "share_network_id": fake_net_id,
            }
        }

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create,
                          self.request, body)

        self.assertFalse(mock_api_call.called)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_source_sg_snapshot_id(self):
        fake_snap_id = uuidutils.generate_uuid()
        fake_share_group, expected_group = self._get_fake_share_group(
            source_share_group_snapshot_id=fake_snap_id)
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_share_group))

        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id,
            }
        }

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, share_group_type_id=self.fake_share_group_type['id'],
            source_share_group_snapshot_id=fake_snap_id)
        self.assertEqual(expected_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_share_network_id(self):
        fake_net_id = uuidutils.generate_uuid()
        fake_group, expected_group = self._get_fake_share_group(
            share_network_id=fake_net_id)

        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_group))
        body = {
            "share_group": {
                "share_network_id": fake_net_id,
            }
        }

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, share_network_id=fake_net_id,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=mock.ANY)
        self.assertEqual(expected_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_sg_create_no_default_share_type_with_share_group_snapshot(self):
        fake_snap_id = uuidutils.generate_uuid()
        fake, expected = self._get_fake_share_group()
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=None))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake))
        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id,
            }
        }

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, share_group_type_id=self.fake_share_group_type['id'],
            source_share_group_snapshot_id=fake_snap_id)
        self.assertEqual(expected, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_name_and_description(self):
        fake_name = 'fake_name'
        fake_description = 'fake_description'
        fake_group, expected_group = self._get_fake_share_group(
            name=fake_name, description=fake_description)
        self.mock_object(share_types, 'get_default_share_type',
                         mock.Mock(return_value=self.fake_share_type))
        self.mock_object(self.controller.share_group_api, 'create',
                         mock.Mock(return_value=fake_group))
        body = {
            "share_group": {
                "name": fake_name,
                "description": fake_description
            }
        }

        res_dict = self.controller.create(self.request, body)

        self.controller.share_group_api.create.assert_called_once_with(
            self.context, name=fake_name, description=fake_description,
            share_group_type_id=self.fake_share_group_type['id'],
            share_type_ids=[self.fake_share_type['id']])
        self.assertEqual(expected_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_invalid_body(self):
        body = {"not_group": {}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_group_create_invalid_body_share_types_and_source_group_snapshot(
            self):
        body = {
            "share_group": {
                "share_types": [],
                "source_share_group_snapshot_id": "",
            }
        }
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_source_group_snapshot_not_in_available(self):
        fake_snap_id = uuidutils.generate_uuid()
        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id,
            }
        }
        self.mock_object(self.controller.share_group_api, 'create', mock.Mock(
            side_effect=exception.InvalidShareGroupSnapshot(reason='blah')))

        self.assertRaises(
            webob.exc.HTTPConflict, self.controller.create, self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_source_group_snapshot_does_not_exist(self):
        fake_snap_id = uuidutils.generate_uuid()
        body = {
            "share_group": {"source_share_group_snapshot_id": fake_snap_id}
        }
        self.mock_object(
            self.controller.share_group_api, 'create',
            mock.Mock(side_effect=exception.ShareGroupSnapshotNotFound(
                share_group_snapshot_id=fake_snap_id)))

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.create, self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_source_group_snapshot_not_a_uuid(self):
        fake_snap_id = "Not a uuid"
        body = {
            "share_group": {
                "source_share_group_snapshot_id": fake_snap_id,
            }
        }

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_share_network_id_not_a_uuid(self):
        fake_net_id = "Not a uuid"
        body = {"share_group": {"share_network_id": fake_net_id}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_invalid_body_share_types_not_a_list(self):
        body = {"share_group": {"share_types": ""}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_invalid_body_invalid_field(self):
        body = {"share_group": {"unknown_field": ""}}

        exc = self.assertRaises(webob.exc.HTTPBadRequest,
                                self.controller.create,
                                self.request, body)

        self.assertIn('unknown_field', str(exc))
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_invalid_share_types_field(self):
        body = {"share_group": {"share_types": 'iamastring'}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    def test_share_group_create_with_invalid_share_types_field_not_uuids(self):
        body = {"share_group": {"share_types": ['iamastring']}}

        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          self.request, body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'create')

    @ddt.data({'microversion': '2.34', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_update_with_name_and_description(
            self, microversion, experimental):
        fake_name = 'fake_name'
        fake_description = 'fake_description'
        fake_group, expected_group = self._get_fake_share_group(
            name=fake_name, description=fake_description)
        self.mock_object(self.controller.share_group_api, 'get',
                         mock.Mock(return_value=fake_group))
        self.mock_object(self.controller.share_group_api, 'update',
                         mock.Mock(return_value=fake_group))
        req, req_context = self._get_fake_custom_request_and_context(
            microversion, experimental)
        body = {
            "share_group": {
                "name": fake_name,
                "description": fake_description,
            }
        }

        res_dict = self.controller.update(req, fake_group['id'], body)

        self.controller.share_group_api.update.assert_called_once_with(
            req_context, fake_group,
            {"name": fake_name, "description": fake_description})
        self.assertEqual(expected_group, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'update')

    def test_share_group_update_group_not_found(self):
        body = {"share_group": {}}
        self.mock_object(self.controller.share_group_api, 'get',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.update,
                          self.request, 'fake_id', body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'update')

    def test_share_group_update_invalid_body(self):
        body = {"not_group": {}}

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update,
                          self.request, 'fake_id', body)

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'update')

    def test_share_group_update_invalid_body_invalid_field(self):
        body = {"share_group": {"unknown_field": ""}}

        exc = self.assertRaises(webob.exc.HTTPBadRequest,
                                self.controller.update,
                                self.request, 'fake_id', body)

        self.assertIn('unknown_field', str(exc))
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'update')

    def test_share_group_update_invalid_body_readonly_field(self):
        body = {"share_group": {"share_types": []}}

        exc = self.assertRaises(webob.exc.HTTPBadRequest,
                                self.controller.update,
                                self.request, 'fake_id', body)

        self.assertIn('share_types', str(exc))
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'update')

    @ddt.data({'microversion': '2.31', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_list_index(self, microversion, experimental):
        fake, expected = self._get_fake_simple_share_group()
        self.mock_object(
            share_group_api.API, 'get_all', mock.Mock(return_value=[fake]))
        req, req_context = self._get_fake_custom_request_and_context(
            microversion, experimental)

        res_dict = self.controller.index(req)

        self.assertEqual([expected], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_share_group_list_index_no_groups(self):
        self.mock_object(
            share_group_api.API, 'get_all', mock.Mock(return_value=[]))

        res_dict = self.controller.index(self.request)

        self.assertEqual([], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'get_all')

    def test_share_group_list_index_with_limit(self):
        fake, expected = self._get_fake_simple_share_group()
        fake2, expected2 = self._get_fake_simple_share_group(id="fake_id2")
        self.mock_object(
            share_group_api.API, 'get_all',
            mock.Mock(return_value=[fake, fake2]))
        req = fakes.HTTPRequest.blank(
            '/share-groups?limit=1', version=self.api_version,
            experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict['share_groups']))
        self.assertEqual([expected], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_share_group_list_index_with_limit_and_offset(self):
        fake, expected = self._get_fake_simple_share_group()
        fake2, expected2 = self._get_fake_simple_share_group(
            id="fake_id2")
        self.mock_object(share_group_api.API, 'get_all',
                         mock.Mock(return_value=[fake, fake2]))
        req = fakes.HTTPRequest.blank(
            '/share-groups?limit=1&offset=1', version=self.api_version,
            experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict['share_groups']))
        self.assertEqual([expected2], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_share_group_list_index_with_like_filter(self):
        fake, expected = self._get_fake_simple_share_group(
            name='fake_1', description='fake_ds_1')
        fake2, expected2 = self._get_fake_simple_share_group(
            name='fake_2', description='fake_ds_2')
        self.mock_object(share_group_api.API, 'get_all',
                         mock.Mock(return_value=[fake, fake2]))
        req = fakes.HTTPRequest.blank(
            '/share-groups?name~=fake&description~=fake',
            version='2.36', experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.index(req)

        expected.pop('description')
        expected2.pop('description')
        self.assertEqual(2, len(res_dict['share_groups']))
        self.assertEqual([expected, expected2], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    @ddt.data({'microversion': '2.34', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_list_detail(self, microversion, experimental):
        fake, expected = self._get_fake_share_group()
        self.mock_object(
            share_group_api.API, 'get_all', mock.Mock(return_value=[fake]))
        req, req_context = self._get_fake_custom_request_and_context(
            microversion, experimental)

        res_dict = self.controller.detail(req)

        self.assertEqual([expected], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_share_group_list_detail_no_groups(self):
        self.mock_object(
            share_group_api.API, 'get_all', mock.Mock(return_value=[]))

        res_dict = self.controller.detail(self.request)

        self.assertEqual([], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'get_all')

    def test_share_group_list_detail_with_limit(self):
        req = fakes.HTTPRequest.blank('/share-groups?limit=1',
                                      version=self.api_version,
                                      experimental=True)
        req_context = req.environ['manila.context']
        fake_group, expected_group = self._get_fake_share_group(
            ctxt=req_context)
        fake_group2, expected_group2 = self._get_fake_share_group(
            ctxt=req_context, id="fake_id2")
        self.mock_object(share_group_api.API, 'get_all',
                         mock.Mock(return_value=[fake_group, fake_group2]))

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_groups']))
        self.assertEqual([expected_group], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_share_group_list_detail_with_limit_and_offset(self):
        req = fakes.HTTPRequest.blank('/share-groups?limit=1&offset=1',
                                      version=self.api_version,
                                      experimental=True)
        req_context = req.environ['manila.context']
        fake_group, expected_group = self._get_fake_share_group(
            ctxt=req_context)
        fake_group2, expected_group2 = self._get_fake_share_group(
            id="fake_id2", ctxt=req_context)
        self.mock_object(share_group_api.API, 'get_all',
                         mock.Mock(return_value=[fake_group, fake_group2]))

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_groups']))
        self.assertEqual([expected_group2], res_dict['share_groups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    @ddt.data({'microversion': '2.31', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_delete(self, microversion, experimental):
        fake_group, expected_group = self._get_fake_share_group()
        self.mock_object(share_group_api.API, 'get',
                         mock.Mock(return_value=fake_group))
        self.mock_object(share_group_api.API, 'delete')
        req, req_context = self._get_fake_custom_request_and_context(
            microversion, experimental)

        res = self.controller.delete(req, fake_group['id'])

        self.assertEqual(202, res.status_code)
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'delete')

    def test_share_group_delete_group_not_found(self):
        fake_group, expected_group = self._get_fake_share_group()
        self.mock_object(share_group_api.API, 'get',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(webob.exc.HTTPNotFound, self.controller.delete,
                          self.request, fake_group['id'])
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'delete')

    def test_share_group_delete_in_conflicting_status(self):
        fake, expected = self._get_fake_share_group()
        self.mock_object(
            share_group_api.API, 'get', mock.Mock(return_value=fake))
        self.mock_object(share_group_api.API, 'delete', mock.Mock(
            side_effect=exception.InvalidShareGroup(reason='blah')))

        self.assertRaises(
            webob.exc.HTTPConflict,
            self.controller.delete, self.request, fake['id'])

        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'delete')

    @ddt.data({'microversion': '2.34', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test_share_group_show(self, microversion, experimental):
        fake, expected = self._get_fake_share_group()
        self.mock_object(
            share_group_api.API, 'get', mock.Mock(return_value=fake))
        req = fakes.HTTPRequest.blank(
            '/share-groupss/%s' % fake['id'], version=microversion,
            experimental=experimental)
        req_context = req.environ['manila.context']

        res_dict = self.controller.show(req, fake['id'])

        self.assertEqual(expected, res_dict['share_group'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get')

    def test_share_group_show_as_admin(self):
        req = fakes.HTTPRequest.blank(
            '/share-groupss/my_group_id', version=self.api_version,
            experimental=True)
        admin_context = req.environ['manila.context'].elevated()
        req.environ['manila.context'] = admin_context
        fake_group, expected_group = self._get_fake_share_group(
            ctxt=admin_context, id='my_group_id')
        self.mock_object(share_group_api.API, 'get',
                         mock.Mock(return_value=fake_group))

        res_dict = self.controller.show(req, fake_group['id'])

        self.assertEqual(expected_group, res_dict['share_group'])
        self.assertIsNotNone(res_dict['share_group']['share_server_id'])
        self.mock_policy_check.assert_called_once_with(
            admin_context, self.resource_name, 'get')

    def test_share_group_show_group_not_found(self):
        req = fakes.HTTPRequest.blank(
            '/share-groupss/myfakegroup', version=self.api_version,
            experimental=True)
        req_context = req.environ['manila.context']
        fake, expected = self._get_fake_share_group(
            ctxt=req_context, id='myfakegroup')
        self.mock_object(share_group_api.API, 'get',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(
            webob.exc.HTTPNotFound, self.controller.show, req, fake['id'])

        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get')

    @ddt.data({'microversion': '2.31', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test__reset_status_call(self, microversion, experimental):
        self.mock_object(self.controller, '_reset_status')
        req, _junk = self._get_fake_custom_request_and_context(
            microversion, experimental)
        sg_id = 'fake'
        body = {'reset_status': {'status': constants.STATUS_ERROR}}

        self.controller.share_group_reset_status(req, sg_id, body)
        self.controller._reset_status.assert_called_once_with(req, sg_id, body)

    @ddt.data(*fakes.fixture_reset_status_with_different_roles)
    @ddt.unpack
    def test_share_groups_reset_status_with_different_roles(
            self, role, valid_code, valid_status, version):
        ctxt = self._get_context(role)
        share_group, req = self._setup_share_group_data()

        action_name = 'reset_status'
        body = {action_name: {'status': constants.STATUS_ERROR}}
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers['X-Openstack-Manila-Api-Version'] = self.api_version
        req.environ['manila.context'] = ctxt

        with mock.patch.object(
                policy, 'check_policy', fakes.mock_fake_admin_check):
            resp = req.get_response(fakes.app())

        # validate response code and model status
        self.assertEqual(valid_code, resp.status_int)

        actual_model = db.share_group_get(ctxt, share_group['id'])
        self.assertEqual(valid_status, actual_model['status'])

    @ddt.data(*fakes.fixture_force_delete_with_different_roles)
    @ddt.unpack
    def test_share_group_force_delete_with_different_roles(self, role,
                                                           resp_code, version):
        ctxt = self._get_context(role)
        share_group, req = self._setup_share_group_data()
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        action_name = 'force_delete'
        body = {action_name: {}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers['X-Openstack-Manila-Api-Version'] = self.api_version
        req.environ['manila.context'] = ctxt

        with mock.patch.object(
                policy, 'check_policy', fakes.mock_fake_admin_check):
            resp = req.get_response(fakes.app())

        # validate response
        self.assertEqual(resp_code, resp.status_int)

    @ddt.data({'microversion': '2.31', 'experimental': True},
              {'microversion': SG_GRADUATION_VERSION, 'experimental': False})
    @ddt.unpack
    def test__force_delete_call(self, microversion, experimental):
        self.mock_object(self.controller, '_force_delete')
        req, _junk = self._get_fake_custom_request_and_context(
            microversion, experimental)
        sg_id = 'fake'
        body = {'force_delete': {}}

        self.controller.share_group_force_delete(req, sg_id, body)
        self.controller._force_delete.assert_called_once_with(req, sg_id, body)



# Copyright (c) 2014 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""EMC Share Driver Plugin Framework."""
from stevedore import extension


class EMCPluginManager(object):
    def __init__(self, namespace):
        self.namespace = namespace

        self.extension_manager = extension.ExtensionManager(namespace)

    def load_plugin(self, name, *args, **kwargs):
        for ext in self.extension_manager.extensions:
            if ext.name == name:
                storage_conn = ext.plugin(*args, **kwargs)
                return storage_conn

        return None



# Copyright 2018 Huawei Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The share access rule metadata api."""

import webob

from manila.api.openstack import wsgi
from manila.api.views import share_accesses as share_access_views
from manila import db
from manila import exception
from manila.i18n import _
from manila import share


class ShareAccessMetadataController(wsgi.Controller):
    """The Share access rule metadata API V2 controller."""

    resource_name = 'share_access_metadata'
    _view_builder_class = share_access_views.ViewBuilder

    def __init__(self):
        super(ShareAccessMetadataController, self).__init__()
        self.share_api = share.API()

    @wsgi.Controller.api_version('2.45')
    @wsgi.Controller.authorize
    def update(self, req, access_id, body=None):
        context = req.environ['manila.context']
        if not self.is_valid_body(body, 'metadata'):
            raise webob.exc.HTTPBadRequest()

        metadata = body['metadata']
        md = self._update_share_access_metadata(context, access_id, metadata)
        return self._view_builder.view_metadata(req, md)

    @wsgi.Controller.api_version('2.45')
    @wsgi.Controller.authorize
    @wsgi.response(200)
    def delete(self, req, access_id, key):
        """Deletes an existing access metadata."""
        context = req.environ['manila.context']
        self._assert_access_exists(context, access_id)
        try:
            db.share_access_metadata_delete(context, access_id, key)
        except exception.ShareAccessMetadataNotFound as error:
            raise webob.exc.HTTPNotFound(explanation=error.msg)

    def _update_share_access_metadata(self, context, access_id, metadata):
        self._assert_access_exists(context, access_id)
        try:
            return self.share_api.update_share_access_metadata(
                context, access_id, metadata)
        except (ValueError, AttributeError):
            msg = _("Malformed request body")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        except exception.InvalidMetadata as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)

        except exception.InvalidMetadataSize as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)

    def _assert_access_exists(self, context, access_id):
        try:
            self.share_api.access_get(context, access_id)
        except exception.NotFound as ex:
            raise webob.exc.HTTPNotFound(explanation=ex.msg)


def create_resource():
    return wsgi.Resource(ShareAccessMetadataController())



# Copyright (c) 2018 Huawei Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
import webob

from manila.api.v2 import share_access_metadata
from manila.api.v2 import share_accesses
from manila import context
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils
from oslo_utils import uuidutils


@ddt.ddt
class ShareAccessesMetadataAPITest(test.TestCase):

    def _get_request(self, version="2.45", use_admin_context=True):
        req = fakes.HTTPRequest.blank(
            '/v2/share-access-rules',
            version=version, use_admin_context=use_admin_context)
        return req

    def setUp(self):
        super(ShareAccessesMetadataAPITest, self).setUp()
        self.controller = (
            share_access_metadata.ShareAccessMetadataController())
        self.access_controller = (
            share_accesses.ShareAccessesController())
        self.resource_name = self.controller.resource_name
        self.admin_context = context.RequestContext('admin', 'fake', True)
        self.member_context = context.RequestContext('fake', 'fake')
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.share = db_utils.create_share()
        self.access = db_utils.create_share_access(
            id=uuidutils.generate_uuid(),
            share_id=self.share['id'])

    @ddt.data({'body': {'metadata': {'key1': 'v1'}}},
              {'body': {'metadata': {'test_key1': 'test_v1'}}},
              {'body': {'metadata': {'key1': 'v2'}}})
    @ddt.unpack
    def test_update_metadata(self, body):
        url = self._get_request()
        update = self.controller.update(url, self.access['id'], body=body)
        self.assertEqual(body, update)

        show_result = self.access_controller.show(url, self.access['id'])

        self.assertEqual(1, len(show_result))
        self.assertIn(self.access['id'], show_result['access']['id'])
        self.assertEqual(body['metadata'], show_result['access']['metadata'])

    def test_delete_metadata(self):
        body = {'metadata': {'test_key3': 'test_v3'}}
        url = self._get_request()
        self.controller.update(url, self.access['id'], body=body)

        self.controller.delete(url, self.access['id'], 'test_key3')
        show_result = self.access_controller.show(url, self.access['id'])

        self.assertEqual(1, len(show_result))
        self.assertIn(self.access['id'], show_result['access']['id'])
        self.assertNotIn('test_key3', show_result['access']['metadata'])

    def test_update_access_metadata_with_access_id_not_found(self):
        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.update,
            self._get_request(), 'not_exist_access_id',
            {'metadata': {'key1': 'v1'}})

    def test_update_access_metadata_with_body_error(self):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            self._get_request(), self.access['id'],
            {'metadata_error': {'key1': 'v1'}})

    @ddt.data({'metadata': {'key1': 'v1', 'key2': None}},
              {'metadata': {None: 'v1', 'key2': 'v2'}},
              {'metadata': {'k' * 256: 'v2'}},
              {'metadata': {'key1': 'v' * 1024}})
    @ddt.unpack
    def test_update_metadata_with_invalid_metadata(self, metadata):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            self._get_request(), self.access['id'],
            {'metadata': metadata})

    def test_delete_access_metadata_not_found(self):
        body = {'metadata': {'test_key_exist': 'test_v_exsit'}}
        update = self.controller.update(
            self._get_request(), self.access['id'], body=body)
        self.assertEqual(body, update)
        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.delete,
            self._get_request(), self.access['id'], 'key1')

    @ddt.data('1.0', '2.0', '2.8', '2.44')
    def test_update_metadata_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.update,
            self._get_request(version=version), self.access['id'],
            {'metadata': {'key1': 'v1'}})

    @ddt.data('1.0', '2.0', '2.43')
    def test_delete_metadata_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.delete,
            self._get_request(version=version), self.access['id'], 'key1')



# Copyright 2019 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log

from manila.scheduler.filters import base_host
from manila.share import utils as share_utils

LOG = log.getLogger(__name__)


class CreateFromSnapshotFilter(base_host.BaseHostFilter):
    """CreateFromSnapshotFilter filters hosts based on replication_domain."""

    def host_passes(self, host_state, filter_properties):
        """Return True if new share's host is compatible with snapshot's host.

        Design of this filter:

            - Creating shares from snapshots in another pool or backend needs
              to match with one of the below conditions:
              - The backend of the new share must be the same as its parent
                snapshot.
              - Both new share and snapshot are in the same replication_domain
        """
        snapshot_id = filter_properties.get('request_spec', {}).get(
            'snapshot_id')
        snapshot_host = filter_properties.get(
            'request_spec', {}).get('snapshot_host')

        if None in [snapshot_id, snapshot_host]:
            # NOTE(silvacarlose): if the request does not contain a snapshot_id
            # or a snapshot_host, the user is not creating a share from a
            # snapshot and we don't need to filter out the host.
            return True

        snapshot_backend = share_utils.extract_host(snapshot_host, 'backend')
        snapshot_rep_domain = filter_properties.get('replication_domain')

        host_backend = share_utils.extract_host(host_state.host, 'backend')
        host_rep_domain = host_state.replication_domain

        # Same backend
        if host_backend == snapshot_backend:
            return True
        # Same replication domain
        if snapshot_rep_domain and snapshot_rep_domain == host_rep_domain:
            return True

        msg = ("The parent's snapshot %(snapshot_id)s back end and "
               "replication domain don't match with the back end and "
               "replication domain of the Host %(host)s.")
        kwargs = {
            "snapshot_id": snapshot_id,
            "host": host_state.host
        }
        LOG.debug(msg, kwargs)
        return False



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Transform statuses to lowercase

Revision ID: 3db9992c30f3
Revises: 533646c7af38
Create Date: 2015-05-28 19:30:35.645773

"""

# revision identifiers, used by Alembic.
revision = '3db9992c30f3'
down_revision = '533646c7af38'

from alembic import op
import sqlalchemy as sa

from manila.db.migrations import utils


def upgrade():
    # NOTE(vponomaryov): shares has some statuses as uppercase, so
    # transform them in addition to statuses of share servers.
    for table in ('shares', 'share_servers'):
        _transform_case(table, make_upper=False)


def downgrade():
    # NOTE(vponomaryov): transform share server statuses to uppercase and
    # leave share statuses as is.
    _transform_case('share_servers', make_upper=True)


def _transform_case(table_name, make_upper):
    connection = op.get_bind()
    table = utils.load_table(table_name, connection)
    case = sa.func.upper if make_upper else sa.func.lower

    for row in connection.execute(table.select()):
        op.execute(
            table.update().where(  # pylint: disable=no-value-for-parameter
                table.c.id == row.id
            ).values({'status': case(row.status)})
        )



# Copyright (c) 2012 OpenStack, LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from manila.scheduler.filters import base_host

LOG = log.getLogger(__name__)


class RetryFilter(base_host.BaseHostFilter):
    """Filter out already tried nodes for scheduling purposes."""

    def host_passes(self, host_state, filter_properties):
        """Skip nodes that have already been attempted."""
        retry = filter_properties.get('retry')
        if not retry:
            # Re-scheduling is disabled
            LOG.debug("Re-scheduling is disabled")
            return True

        hosts = retry.get('hosts', [])
        host = host_state.host

        passes = host not in hosts
        pass_msg = "passes" if passes else "fails"

        LOG.debug("Host %(host)s %(pass_msg)s.  Previously tried hosts: "
                  "%(hosts)s",
                  {"host": host, "pass_msg": pass_msg, "hosts": hosts})

        # Host passes if it's not in the list of previously attempted hosts:
        return passes



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import ipaddress
import json
import re
import socket
import sys

from oslo_config import cfg
from oslo_config import types
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import units

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.message import api as message_api
from manila.message import message_field
from manila.share import driver
from manila.share.drivers import ganesha
from manila.share.drivers.ganesha import utils as ganesha_utils
from manila.share.drivers import helpers as driver_helpers

rados = None
json_command = None
ceph_default_target = None


def setup_rados():
    global rados
    if not rados:
        try:
            rados = importutils.import_module('rados')
        except ImportError:
            raise exception.ShareBackendException(
                _("rados python module is not installed"))


def setup_json_command():
    global json_command
    if not json_command:
        try:
            json_command = importutils.import_class(
                'ceph_argparse.json_command')
        except ImportError:
            raise exception.ShareBackendException(
                _("ceph_argparse python module is not installed"))


CEPHX_ACCESS_TYPE = "cephx"

# The default Ceph administrative identity
CEPH_DEFAULT_AUTH_ID = "admin"

DEFAULT_VOLUME_MODE = '755'

RADOS_TIMEOUT = 10

LOG = log.getLogger(__name__)

# Clone statuses
CLONE_CREATING = 'creating'
CLONE_FAILED = 'failed'
CLONE_CANCELED = 'canceled'
CLONE_PENDING = 'pending'
CLONE_INPROGRESS = 'in-progress'
CLONE_COMPLETE = 'complete'

cephfs_opts = [
    cfg.StrOpt('cephfs_conf_path',
               default="",
               help="Fully qualified path to the ceph.conf file."),
    cfg.StrOpt('cephfs_cluster_name',
               help="The name of the cluster in use, if it is not "
                    "the default ('ceph')."
               ),
    cfg.StrOpt('cephfs_auth_id',
               default="manila",
               help="The name of the ceph auth identity to use."
               ),
    cfg.StrOpt('cephfs_volume_path_prefix',
               deprecated_for_removal=True,
               deprecated_since='Wallaby',
               deprecated_reason='This option is not used starting with '
                                 'the Nautilus release of Ceph.',
               default="/volumes",
               help="The prefix of the cephfs volume path."
               ),
    cfg.StrOpt('cephfs_protocol_helper_type',
               default="CEPHFS",
               choices=['CEPHFS', 'NFS'],
               ignore_case=True,
               help="The type of protocol helper to use. Default is "
                    "CEPHFS."
               ),
    cfg.BoolOpt('cephfs_ganesha_server_is_remote',
                default=False,
                help="Whether the NFS-Ganesha server is remote to the driver."
                ),
    cfg.HostAddressOpt('cephfs_ganesha_server_ip',
                       help="The IP address of the NFS-Ganesha server."),
    cfg.StrOpt('cephfs_ganesha_server_username',
               default='root',
               help="The username to authenticate as in the remote "
                    "NFS-Ganesha server host."),
    cfg.StrOpt('cephfs_ganesha_path_to_private_key',
               help="The path of the driver host's private SSH key file."),
    cfg.StrOpt('cephfs_ganesha_server_password',
               secret=True,
               help="The password to authenticate as the user in the remote "
                    "Ganesha server host. This is not required if "
                    "'cephfs_ganesha_path_to_private_key' is configured."),
    cfg.ListOpt('cephfs_ganesha_export_ips',
                default='',
                help="List of IPs to export shares. If not supplied, "
                     "then the value of 'cephfs_ganesha_server_ip' "
                     "will be used to construct share export locations."),
    cfg.StrOpt('cephfs_volume_mode',
               default=DEFAULT_VOLUME_MODE,
               help="The read/write/execute permissions mode for CephFS "
                    "volumes, snapshots, and snapshot groups expressed in "
                    "Octal as with linux 'chmod' or 'umask' commands."),
    cfg.StrOpt('cephfs_filesystem_name',
               help="The name of the filesystem to use, if there are "
                    "multiple filesystems in the cluster."),
]


CONF = cfg.CONF
CONF.register_opts(cephfs_opts)


class RadosError(Exception):
    """Something went wrong talking to Ceph with librados"""

    pass


def rados_command(rados_client, prefix=None, args=None, json_obj=False,
                  target=None):
    """Safer wrapper for ceph_argparse.json_command

    Raises error exception instead of relying on caller to check return
    codes.

    Error exception can result from:
    * Timeout
    * Actual legitimate errors
    * Malformed JSON output

    return: If json_obj is True, return the decoded JSON object from ceph,
            or None if empty string returned.
            If json is False, return a decoded string (the data returned by
            ceph command)
    """

    target = target or ceph_default_target

    if args is None:
        args = {}

    argdict = args.copy()
    argdict['format'] = 'json'

    LOG.debug("Invoking ceph_argparse.json_command - rados_client=%(cl)s, "
              "target=%(tg)s, prefix='%(pf)s', argdict=%(ad)s, "
              "timeout=%(to)s.",
              {"cl": rados_client, "tg": target, "pf": prefix, "ad": argdict,
               "to": RADOS_TIMEOUT})

    try:
        ret, outbuf, outs = json_command(rados_client,
                                         target=target,
                                         prefix=prefix,
                                         argdict=argdict,
                                         timeout=RADOS_TIMEOUT)
        if ret != 0:
            raise rados.Error(outs, ret)
        if not json_obj:
            result = outbuf.decode().strip()
        else:
            if outbuf:
                result = json.loads(outbuf.decode().strip())
            else:
                result = None
    except Exception as e:
        msg = _("json_command failed - prefix=%(pfx)s, argdict=%(ad)s - "
                "exception message: %(ex)s." %
                {"pfx": prefix, "ad": argdict, "ex": e})
        raise exception.ShareBackendException(msg)

    return result


class CephFSDriver(driver.ExecuteMixin, driver.GaneshaMixin,
                   driver.ShareDriver):
    """Driver for the Ceph Filesystem."""

    def __init__(self, *args, **kwargs):
        super(CephFSDriver, self).__init__(False, *args, **kwargs)
        self.backend_name = self.configuration.safe_get(
            'share_backend_name') or 'CephFS'

        setup_rados()
        setup_json_command()
        self._rados_client = None
        # name of the filesystem/volume used by the driver
        self._volname = None
        self._ceph_mon_version = None
        self.configuration.append_config_values(cephfs_opts)

        try:
            int(self.configuration.cephfs_volume_mode, 8)
        except ValueError:
            msg = _("Invalid CephFS volume mode %s")
            raise exception.BadConfigurationException(
                msg % self.configuration.cephfs_volume_mode)

        self._cephfs_volume_mode = self.configuration.cephfs_volume_mode
        self.ipv6_implemented = True

    def do_setup(self, context):
        if self.configuration.cephfs_protocol_helper_type.upper() == "CEPHFS":
            protocol_helper_class = getattr(
                sys.modules[__name__], 'NativeProtocolHelper')
        else:
            protocol_helper_class = getattr(
                sys.modules[__name__], 'NFSProtocolHelper')

        self.setup_default_ceph_cmd_target()

        self.protocol_helper = protocol_helper_class(
            self._execute,
            self.configuration,
            rados_client=self.rados_client,
            volname=self.volname)

        self.protocol_helper.init_helper()

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        self.protocol_helper.check_for_setup_error()

    def _update_share_stats(self):
        stats = self.rados_client.get_cluster_stats()

        total_capacity_gb = round(stats['kb'] / units.Mi, 2)
        free_capacity_gb = round(stats['kb_avail'] / units.Mi, 2)

        data = {
            'vendor_name': 'Ceph',
            'driver_version': '1.0',
            'share_backend_name': self.backend_name,
            'storage_protocol': self.configuration.safe_get(
                'cephfs_protocol_helper_type'),
            'pools': [
                {
                    'pool_name': 'cephfs',
                    'total_capacity_gb': total_capacity_gb,
                    'free_capacity_gb': free_capacity_gb,
                    'qos': 'False',
                    'reserved_percentage': self.configuration.safe_get(
                        'reserved_share_percentage'),
                    'reserved_snapshot_percentage':
                        self.configuration.safe_get(
                            'reserved_share_from_snapshot_percentage') or
                        self.configuration.safe_get(
                            'reserved_share_percentage'),
                    'dedupe': [False],
                    'compression': [False],
                    'thin_provisioning': [False]
                }
            ],
            'total_capacity_gb': total_capacity_gb,
            'free_capacity_gb': free_capacity_gb,
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
        }
        super(    # pylint: disable=no-member
            CephFSDriver, self)._update_share_stats(data)

    def _to_bytes(self, gigs):
        """Convert a Manila size into bytes.

        Manila uses gibibytes everywhere.

        :param gigs: integer number of gibibytes.
        :return: integer number of bytes.
        """
        return gigs * units.Gi

    def _get_export_locations(self, share):
        """Get the export location for a share.

        :param share: a manila share.
        :return: the export location for a share.
        """

        # get path of FS subvolume/share
        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"]
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        subvolume_path = rados_command(
            self.rados_client, "fs subvolume getpath", argdict)

        return self.protocol_helper.get_export_locations(share, subvolume_path)

    def setup_default_ceph_cmd_target(self):
        global ceph_default_target
        if not ceph_default_target:
            ceph_default_target = ('mon-mgr', )

            try:
                ceph_major_version = self.ceph_mon_version['major']
            except Exception:
                msg = _("Error reading ceph version to set the default "
                        "target. Please check your Ceph backend is reachable.")
                raise exception.ShareBackendException(msg=msg)

            if ceph_major_version == '14':
                ceph_default_target = ('mgr', )
            elif ceph_major_version < '14':
                msg = _("CephFSDriver does not support Ceph "
                        "cluster version less than 14.x (Nautilus)")
                raise exception.ShareBackendException(msg=msg)

    @property
    def ceph_mon_version(self):
        if self._ceph_mon_version:
            return self._ceph_mon_version

        self._ceph_mon_version = {}

        output = rados_command(self.rados_client, "version", target=('mon', ))

        version_str = json.loads(output)["version"]

        p = re.compile(r"ceph version (\d+)\.(\d+)\.(\d+)")
        major, minor, extra = p.match(version_str).groups()
        self._ceph_mon_version['major'] = major
        self._ceph_mon_version['minor'] = minor
        self._ceph_mon_version['extra'] = extra

        return self._ceph_mon_version

    @property
    def rados_client(self):
        if self._rados_client:
            return self._rados_client

        conf_path = self.configuration.safe_get('cephfs_conf_path')
        cluster_name = self.configuration.safe_get('cephfs_cluster_name')
        auth_id = self.configuration.safe_get('cephfs_auth_id')
        self._rados_client = rados.Rados(
            name="client.{0}".format(auth_id),
            clustername=cluster_name,
            conffile=conf_path,
            conf={}
        )

        LOG.info("[%(be)s] Ceph client found, connecting...",
                 {"be": self.backend_name})
        try:
            if self._rados_client.state != "connected":
                self._rados_client.connect()
        except Exception:
            self._rados_client = None
            raise exception.ShareBackendException(
                "[%(be)s] Ceph client failed to connect.",
                {"be": self.backend_name})
        else:
            LOG.info("[%(be)s] Ceph client connection complete.",
                     {"be": self.backend_name})

        return self._rados_client

    @property
    def volname(self):
        # Name of the CephFS volume/filesystem where the driver creates
        # manila entities such as shares, sharegroups, snapshots, etc.
        if self._volname:
            return self._volname

        self._volname = self.configuration.safe_get('cephfs_filesystem_name')
        if not self._volname:
            out = rados_command(
                self.rados_client, "fs volume ls", json_obj=True)
            if len(out) == 1:
                self._volname = out[0]['name']
            else:
                if len(out) > 1:
                    msg = _("Specify Ceph filesystem name using "
                            "'cephfs_filesystem_name' driver option.")
                else:
                    msg = _("No Ceph filesystem found.")
                raise exception.ShareBackendException(msg=msg)

        return self._volname

    def create_share(self, context, share, share_server=None):
        """Create a CephFS volume.

        :param context: A RequestContext.
        :param share: A Share.
        :param share_server: Always None for CephFS native.
        :return: The export locations dictionary.
        """
        requested_proto = share['share_proto'].upper()
        supported_proto = (
            self.configuration.cephfs_protocol_helper_type.upper())
        if (requested_proto != supported_proto):
            msg = _("Share protocol %s is not supported.") % requested_proto
            raise exception.ShareBackendException(msg=msg)
        size = self._to_bytes(share['size'])

        LOG.debug("[%(be)s]: create_share: id=%(id)s, size=%(sz)s, "
                  "group=%(gr)s.",
                  {"be": self.backend_name, "id": share['id'],
                   "sz": share['size'], "gr": share['share_group_id']})

        # create FS subvolume/share
        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "size": size,
            "namespace_isolated": True,
            "mode": self._cephfs_volume_mode,
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        rados_command(self.rados_client, "fs subvolume create", argdict)

        return self._get_export_locations(share)

    def _need_to_cancel_clone(self, share):
        # Is there an ongoing clone operation that needs to be canceled
        # so we can delete the share?
        need_to_cancel_clone = False

        argdict = {
            "vol_name": self.volname,
            "clone_name": share["id"],
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        try:
            status = rados_command(
                self.rados_client, "fs clone status", argdict)
            if status in (CLONE_PENDING, CLONE_INPROGRESS):
                need_to_cancel_clone = True
        except exception.ShareBackendException as e:
            # Trying to get clone status on a regular subvolume is expected
            # to fail.
            if 'not allowed on subvolume' not in str(e).lower():
                raise exception.ShareBackendException(
                    "Failed to remove share.")

        return need_to_cancel_clone

    def delete_share(self, context, share, share_server=None):
        # remove FS subvolume/share
        LOG.debug("[%(be)s]: delete_share: id=%(id)s, group=%(gr)s.",
                  {"be": self.backend_name, "id": share['id'],
                   "gr": share['share_group_id']})

        if self._need_to_cancel_clone(share):
            try:
                argdict = {
                    "vol_name": self.volname,
                    "clone_name": share["id"],
                    "force": True,
                }
                if share['share_group_id'] is not None:
                    argdict.update({"group_name": share["share_group_id"]})

                rados_command(self.rados_client, "fs clone cancel", argdict)
            except rados.Error:
                raise exception.ShareBackendException(
                    "Failed to cancel clone operation.")

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            # We want to clean up the share even if the subvolume is
            # not in a good state.
            "force": True,
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        rados_command(self.rados_client, "fs subvolume rm", argdict)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        return self.protocol_helper.update_access(
            context, share, access_rules, add_rules, delete_rules,
            share_server=share_server)

    def ensure_share(self, context, share, share_server=None):
        # Creation is idempotent
        return self.create_share(context, share, share_server)

    def extend_share(self, share, new_size, share_server=None):
        # resize FS subvolume/share
        LOG.debug("[%(be)s]: extend_share: share=%(id)s, size=%(sz)s.",
                  {"be": self.backend_name, "id": share['id'],
                   "sz": new_size})

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "new_size": self._to_bytes(new_size),
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        LOG.debug("extend_share {id} {size}".format(
            id=share['id'], size=new_size))

        rados_command(self.rados_client, "fs subvolume resize", argdict)

    def shrink_share(self, share, new_size, share_server=None):
        # resize FS subvolume/share
        LOG.debug("[%(be)s]: shrink_share: share=%(id)s, size=%(sz)s.",
                  {"be": self.backend_name, "id": share['id'],
                   "sz": new_size})

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "new_size": self._to_bytes(new_size),
            "no_shrink": True,
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        try:
            rados_command(self.rados_client, "fs subvolume resize", argdict)
        except exception.ShareBackendException as e:
            if 'would be lesser than' in str(e).lower():
                raise exception.ShareShrinkingPossibleDataLoss(
                    share_id=share['id'])
            raise

    def create_snapshot(self, context, snapshot, share_server=None):
        # create a FS snapshot
        LOG.debug("[%(be)s]: create_snapshot: original share=%(id)s, "
                  "snapshot=%(sn)s.",
                  {"be": self.backend_name, "id": snapshot['share_id'],
                   "sn": snapshot['id']})

        argdict = {
            "vol_name": self.volname,
            "sub_name": snapshot["share_id"],
            "snap_name": "_".join([snapshot["snapshot_id"], snapshot["id"]]),
        }

        rados_command(
            self.rados_client, "fs subvolume snapshot create", argdict)

    def delete_snapshot(self, context, snapshot, share_server=None):
        # delete a FS snapshot
        LOG.debug("[%(be)s]: delete_snapshot: snapshot=%(id)s.",
                  {"be": self.backend_name, "id": snapshot['id']})
        argdict = {
            "vol_name": self.volname,
            "sub_name": snapshot["share_id"],
            "snap_name": '_'.join([snapshot['snapshot_id'], snapshot['id']]),
            "force": True,
        }

        rados_command(self.rados_client, "fs subvolume snapshot rm", argdict)

    def create_share_group(self, context, sg_dict, share_server=None):
        # delete a FS group
        LOG.debug("[%(be)s]: create_share_group: share_group=%(id)s.",
                  {"be": self.backend_name, "id": sg_dict['id']})

        argdict = {
            "vol_name": self.volname,
            "group_name": sg_dict['id'],
            "mode": self._cephfs_volume_mode,
        }

        rados_command(self.rados_client, "fs subvolumegroup create", argdict)

    def delete_share_group(self, context, sg_dict, share_server=None):
        # delete a FS group
        LOG.debug("[%(be)s]: delete_share_group: share_group=%(id)s.",
                  {"be": self.backend_name, "id": sg_dict['id']})

        argdict = {
            "vol_name": self.volname,
            "group_name": sg_dict['id'],
            "force": True,
        }

        rados_command(self.rados_client, "fs subvolumegroup rm", argdict)

    def delete_share_group_snapshot(self, context, snap_dict,
                                    share_server=None):
        # delete a FS group snapshot
        LOG.debug("[%(be)s]: delete_share_group_snapshot: "
                  "share_group=%(sg_id)s, snapshot=%(sn)s.",
                  {"be": self.backend_name, "sg_id": snap_dict['id'],
                   "sn": snap_dict["share_group_id"]})

        argdict = {
            "vol_name": self.volname,
            "group_name": snap_dict["share_group_id"],
            "snap_name": snap_dict["id"],
            "force": True,
        }

        rados_command(
            self.rados_client, "fs subvolumegroup snapshot rm", argdict)

        return None, []

    def create_share_group_snapshot(self, context, snap_dict,
                                    share_server=None):
        # create a FS group snapshot
        LOG.debug("[%(be)s]: create_share_group_snapshot: share_group=%(id)s, "
                  "snapshot=%(sn)s.",
                  {"be": self.backend_name, "id": snap_dict['share_group_id'],
                   "sn": snap_dict["id"]})

        msg = _("Share group snapshot feature is no longer supported in "
                "mainline CephFS (existing group snapshots can still be "
                "listed and deleted).")
        raise exception.ShareBackendException(msg=msg)

    def _get_clone_status(self, share):
        """Check the status of a newly cloned share."""
        argdict = {
            "vol_name": self.volname,
            "clone_name": share["id"]
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        out = rados_command(self.rados_client,
                            "fs clone status", argdict, True)
        return out['status']['state']

    def _update_create_from_snapshot_status(self, share):
        updates = {
            'status': constants.STATUS_ERROR,
            'progress': None,
            'export_locations': []
        }
        status = self._get_clone_status(share)
        if status == CLONE_COMPLETE:
            updates['status'] = constants.STATUS_AVAILABLE
            updates['progress'] = '100%'
            updates['export_locations'] = self._get_export_locations(share)
        elif status in (CLONE_PENDING, CLONE_INPROGRESS):
            updates['status'] = constants.STATUS_CREATING_FROM_SNAPSHOT
        else:
            # error if clone operation is not progressing or completed
            raise exception.ShareBackendException(
                "rados client clone of snapshot [%(sn)s}] to new "
                "share [%(shr)s}] did not complete successfully." %
                {"sn": share["snapshot_id"], "shr": share["id"]})
        return updates

    def get_share_status(self, share, share_server=None):
        """Returns the current status for a share.

        :param share: a manila share.
        :param share_server: a manila share server (not currently supported).
        :returns: manila share status.
        """

        if share['status'] != constants.STATUS_CREATING_FROM_SNAPSHOT:
            LOG.warning("Caught an unexpected share status '%s' during share "
                        "status update routine. Skipping.", share['status'])
            return
        return self._update_create_from_snapshot_status(share)

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Create a CephFS subvolume from a snapshot"""

        LOG.debug("[%(be)s]: create_share_from_snapshot: id=%(id)s, "
                  "snapshot=%(sn)s, size=%(sz)s, group=%(gr)s.",
                  {"be": self.backend_name, "id": share['id'],
                   "sn": snapshot['id'], "sz": share['size'],
                   "gr": share['share_group_id']})

        argdict = {
            "vol_name": self.volname,
            "sub_name": parent_share["id"],
            "snap_name": '_'.join([snapshot["snapshot_id"], snapshot["id"]]),
            "target_sub_name": share["id"]
        }
        if share['share_group_id'] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        rados_command(
            self.rados_client, "fs subvolume snapshot clone", argdict)

        return self._update_create_from_snapshot_status(share)

    def __del__(self):
        if self._rados_client:
            LOG.info("[%(be)s] Ceph client disconnecting...",
                     {"be": self.backend_name})
            self._rados_client.shutdown()
            self._rados_client = None
            LOG.info("[%(be)s] Ceph client disconnected",
                     {"be": self.backend_name})

    def get_configured_ip_versions(self):
        return self.protocol_helper.get_configured_ip_versions()


class NativeProtocolHelper(ganesha.NASHelperBase):
    """Helper class for native CephFS protocol"""

    supported_access_types = (CEPHX_ACCESS_TYPE, )
    supported_access_levels = (constants.ACCESS_LEVEL_RW,
                               constants.ACCESS_LEVEL_RO)

    def __init__(self, execute, config, **kwargs):
        self.rados_client = kwargs.pop('rados_client')
        self.volname = kwargs.pop('volname')
        self.message_api = message_api.API()
        super(NativeProtocolHelper, self).__init__(execute, config,
                                                   **kwargs)

    def _init_helper(self):
        pass

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        return

    def get_mon_addrs(self):
        result = []
        mon_map = rados_command(self.rados_client, "mon dump", json_obj=True,
                                target=('mon', ))
        for mon in mon_map['mons']:
            ip_port = mon['addr'].split("/")[0]
            result.append(ip_port)

        return result

    def get_export_locations(self, share, subvolume_path):
        # To mount this you need to know the mon IPs and the path to the volume
        mon_addrs = self.get_mon_addrs()

        export_location = "{addrs}:{path}".format(
            addrs=",".join(mon_addrs),
            path=subvolume_path)

        LOG.info("Calculated export location for share %(id)s: %(loc)s",
                 {"id": share['id'], "loc": export_location})

        return {
            'path': export_location,
            'is_admin_only': False,
            'metadata': {},
        }

    def _allow_access(self, context, share, access, share_server=None):
        if access['access_type'] != CEPHX_ACCESS_TYPE:
            raise exception.InvalidShareAccessType(type=access['access_type'])

        ceph_auth_id = access['access_to']

        # We need to check here rather than the API or Manila Client to see
        # if the ceph_auth_id is the same as the one specified for Manila's
        # usage. This is due to the fact that the API and the Manila client
        # cannot read the contents of the Manila configuration file. If it
        # is the same, we need to error out.
        if ceph_auth_id == CONF.cephfs_auth_id:
            error_message = (_('Ceph authentication ID %s must be different '
                             'than the one the Manila service uses.') %
                             ceph_auth_id)
            raise exception.InvalidShareAccess(reason=error_message)

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "auth_id": ceph_auth_id,
            "tenant_id": share["project_id"],
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        readonly = access['access_level'] == constants.ACCESS_LEVEL_RO

        if readonly:
            argdict.update({"access_level": "r"})
        else:
            argdict.update({"access_level": "rw"})

        try:
            auth_result = rados_command(
                self.rados_client, "fs subvolume authorize", argdict)
        except exception.ShareBackendException as e:
            if 'not allowed' in str(e).lower():
                msg = ("Access to client %(client)s is not allowed. "
                       "Reason: %(reason)s")
                msg_payload = {'client': ceph_auth_id, 'reason': e}
                raise exception.InvalidShareAccess(
                    reason=msg % msg_payload)
            raise

        return auth_result

    def _deny_access(self, context, share, access, share_server=None):
        if access['access_type'] != CEPHX_ACCESS_TYPE:
            LOG.warning("Invalid access type '%(type)s', "
                        "ignoring in deny.",
                        {"type": access['access_type']})
            return

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "auth_id": access['access_to']
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        rados_command(self.rados_client, "fs subvolume deauthorize", argdict)
        rados_command(self.rados_client, "fs subvolume evict", argdict)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        access_updates = {}

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        if not (add_rules or delete_rules):  # recovery/maintenance mode
            add_rules = access_rules

            existing_auths = None

            existing_auths = rados_command(
                self.rados_client, "fs subvolume authorized_list",
                argdict, json_obj=True)

            if existing_auths:
                existing_auth_ids = set()
                for rule in range(len(existing_auths)):
                    for cephx_id in existing_auths[rule]:
                        existing_auth_ids.add(cephx_id)
                want_auth_ids = set(
                    [rule['access_to'] for rule in add_rules])
                delete_auth_ids = existing_auth_ids.difference(
                    want_auth_ids)
                delete_auth_ids_list = delete_auth_ids
                for delete_auth_id in delete_auth_ids_list:
                    delete_rules.append(
                        {
                            'access_to': delete_auth_id,
                            'access_type': CEPHX_ACCESS_TYPE,
                        })

        # During recovery mode, re-authorize share access for auth IDs that
        # were already granted access by the backend. Do this to fetch their
        # access keys and ensure that after recovery, manila and the Ceph
        # backend are in sync.
        for rule in add_rules:
            try:
                access_key = self._allow_access(context, share, rule)
            except (exception.InvalidShareAccessLevel,
                    exception.InvalidShareAccessType):
                self.message_api.create(
                    context,
                    message_field.Action.UPDATE_ACCESS_RULES,
                    share['project_id'],
                    resource_type=message_field.Resource.SHARE,
                    resource_id=share['share_id'],
                    detail=message_field.Detail.UNSUPPORTED_CLIENT_ACCESS)
                log_args = {'id': rule['access_id'],
                            'access_level': rule['access_level'],
                            'access_to': rule['access_to']}
                LOG.exception("Failed to provide %(access_level)s access to "
                              "%(access_to)s (Rule ID: %(id)s). Setting rule "
                              "to 'error' state.", log_args)
                access_updates.update({rule['access_id']: {'state': 'error'}})
            except exception.InvalidShareAccess:
                self.message_api.create(
                    context,
                    message_field.Action.UPDATE_ACCESS_RULES,
                    share['project_id'],
                    resource_type=message_field.Resource.SHARE,
                    resource_id=share['share_id'],
                    detail=message_field.Detail.FORBIDDEN_CLIENT_ACCESS)
                log_args = {'id': rule['access_id'],
                            'access_level': rule['access_level'],
                            'access_to': rule['access_to']}
                LOG.exception("Failed to provide %(access_level)s access to "
                              "%(access_to)s (Rule ID: %(id)s). Setting rule "
                              "to 'error' state.", log_args)
                access_updates.update({rule['access_id']: {'state': 'error'}})
            else:
                access_updates.update({
                    rule['access_id']: {'access_key': access_key},
                })

        for rule in delete_rules:
            self._deny_access(context, share, rule)

        return access_updates

    def get_configured_ip_versions(self):
        return [4]


class NFSProtocolHelper(ganesha.GaneshaNASHelper2):

    shared_data = {}
    supported_protocols = ('NFS',)

    def __init__(self, execute, config_object, **kwargs):
        if config_object.cephfs_ganesha_server_is_remote:
            execute = ganesha_utils.SSHExecutor(
                config_object.cephfs_ganesha_server_ip, 22, None,
                config_object.cephfs_ganesha_server_username,
                password=config_object.cephfs_ganesha_server_password,
                privatekey=config_object.cephfs_ganesha_path_to_private_key)
        else:
            execute = ganesha_utils.RootExecutor(execute)

        self.ganesha_host = config_object.cephfs_ganesha_server_ip
        if not self.ganesha_host:
            self.ganesha_host = socket.gethostname()
            LOG.info("NFS-Ganesha server's location defaulted to driver's "
                     "hostname: %s", self.ganesha_host)

        super(NFSProtocolHelper, self).__init__(execute, config_object,
                                                **kwargs)

        if not hasattr(self, 'rados_client'):
            self.rados_client = kwargs.pop('rados_client')
        if not hasattr(self, 'volname'):
            self.volname = kwargs.pop('volname')
        self.export_ips = config_object.cephfs_ganesha_export_ips
        if not self.export_ips:
            self.export_ips = [self.ganesha_host]
        self.configured_ip_versions = set()
        self.config = config_object

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        host_address_obj = types.HostAddress()
        for export_ip in self.config.cephfs_ganesha_export_ips:
            try:
                host_address_obj(export_ip)
            except ValueError:
                msg = (_("Invalid list member of 'cephfs_ganesha_export_ips' "
                         "option supplied %s -- not a valid IP address or "
                         "hostname.") % export_ip)
                raise exception.InvalidParameterValue(err=msg)

    def get_export_locations(self, share, subvolume_path):
        export_locations = []
        for export_ip in self.export_ips:
            # Try to escape the export ip. If it fails, means that the
            # `cephfs_ganesha_server_ip` wasn't possibly set and the used
            # address is the hostname
            try:
                server_address = driver_helpers.escaped_address(export_ip)
            except ValueError:
                server_address = export_ip

            export_path = "{server_address}:{mount_path}".format(
                server_address=server_address, mount_path=subvolume_path)

            LOG.info("Calculated export path for share %(id)s: %(epath)s",
                     {"id": share['id'], "epath": export_path})
            export_location = {
                'path': export_path,
                'is_admin_only': False,
                'metadata': {},
            }
            export_locations.append(export_location)
        return export_locations

    def _default_config_hook(self):
        """Callback to provide default export block."""
        dconf = super(NFSProtocolHelper, self)._default_config_hook()
        conf_dir = ganesha_utils.path_from(__file__, "conf")
        ganesha_utils.patch(dconf, self._load_conf_dir(conf_dir))
        return dconf

    def _fsal_hook(self, base, share, access):
        """Callback to create FSAL subblock."""
        ceph_auth_id = ''.join(['ganesha-', share['id']])

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "auth_id": ceph_auth_id,
            "access_level": "rw",
            "tenant_id": share["project_id"],
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        auth_result = rados_command(
            self.rados_client, "fs subvolume authorize", argdict)

        # Restrict Ganesha server's access to only the CephFS subtree or path,
        # corresponding to the manila share, that is to be exported by making
        # Ganesha use Ceph auth IDs with path restricted capabilities to
        # communicate with CephFS.
        return {
            'Name': 'Ceph',
            'User_Id': ceph_auth_id,
            'Secret_Access_Key': auth_result
        }

    def _cleanup_fsal_hook(self, base, share, access):
        """Callback for FSAL specific cleanup after removing an export."""
        ceph_auth_id = ''.join(['ganesha-', share['id']])

        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"],
            "auth_id": ceph_auth_id,
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        rados_command(self.rados_client, "fs subvolume deauthorize", argdict)

    def _get_export_path(self, share):
        """Callback to provide export path."""
        argdict = {
            "vol_name": self.volname,
            "sub_name": share["id"]
        }
        if share["share_group_id"] is not None:
            argdict.update({"group_name": share["share_group_id"]})

        path = rados_command(
            self.rados_client, "fs subvolume getpath", argdict)

        return path

    def _get_export_pseudo_path(self, share):
        """Callback to provide pseudo path."""
        return self._get_export_path(share)

    def get_configured_ip_versions(self):
        if not self.configured_ip_versions:
            try:
                for export_ip in self.export_ips:
                    self.configured_ip_versions.add(
                        ipaddress.ip_address(str(export_ip)).version)
            except Exception:
                # export_ips contained a hostname, safest thing is to
                # claim support for IPv4 and IPv6 address families
                LOG.warning("Setting configured IP versions to [4, 6] since "
                            "a hostname (rather than IP address) was supplied "
                            "in 'cephfs_ganesha_server_ip' or "
                            "in 'cephfs_ganesha_export_ips'.")
                return [4, 6]
        return list(self.configured_ip_versions)



# Copyright 2011 Piston Cloud Computing, Inc.
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test of Policy Engine For Manila."""

import ddt
from oslo_config import cfg
from oslo_policy import policy as common_policy

from manila import context
from manila import exception
from manila import policy
from manila import test

CONF = cfg.CONF


@ddt.ddt
class PolicyTestCase(test.TestCase):
    def setUp(self):
        super(PolicyTestCase, self).setUp()
        rules = [
            common_policy.RuleDefault("true", '@'),
            common_policy.RuleDefault("test:allowed", '@'),
            common_policy.RuleDefault("test:denied", "!"),
            common_policy.RuleDefault("test:my_file",
                                      "role:compute_admin or "
                                      "project_id:%(project_id)s"),
            common_policy.RuleDefault("test:early_and_fail", "! and @"),
            common_policy.RuleDefault("test:early_or_success", "@ or !"),
            common_policy.RuleDefault("test:lowercase_admin",
                                      "role:admin"),
            common_policy.RuleDefault("test:uppercase_admin",
                                      "role:ADMIN"),
        ]
        policy.reset()
        policy.init(suppress_deprecation_warnings=True)
        # before a policy rule can be used, its default has to be registered.
        policy._ENFORCER.register_defaults(rules)
        self.context = context.RequestContext('fake', 'fake', roles=['member'])
        self.target = {}
        self.addCleanup(policy.reset)

    def test_authorize_nonexistent_action_throws(self):
        action = "test:noexist"
        self.assertRaises(common_policy.PolicyNotRegistered, policy.authorize,
                          self.context, action, self.target)

    def test_authorize_bad_action_throws(self):
        action = "test:denied"
        self.assertRaises(exception.PolicyNotAuthorized, policy.authorize,
                          self.context, action, self.target)

    def test_authorize_bad_action_noraise(self):
        action = "test:denied"
        result = policy.authorize(self.context, action, self.target, False)
        self.assertFalse(result)

    def test_authorize_good_action(self):
        action = "test:allowed"
        result = policy.authorize(self.context, action, self.target)
        self.assertTrue(result)

    def test_templatized_authorization(self):
        target_mine = {'project_id': 'fake'}
        target_not_mine = {'project_id': 'another'}
        action = "test:my_file"
        policy.authorize(self.context, action, target_mine)
        self.assertRaises(exception.PolicyNotAuthorized, policy.authorize,
                          self.context, action, target_not_mine)

    def test_early_AND_authorization(self):
        action = "test:early_and_fail"
        self.assertRaises(exception.PolicyNotAuthorized, policy.authorize,
                          self.context, action, self.target)

    def test_early_OR_authorization(self):
        action = "test:early_or_success"
        policy.authorize(self.context, action, self.target)

    def test_ignore_case_role_check(self):
        lowercase_action = "test:lowercase_admin"
        uppercase_action = "test:uppercase_admin"
        admin_context = context.RequestContext('admin',
                                               'fake',
                                               roles=['AdMiN'])
        policy.authorize(admin_context, lowercase_action, self.target)
        policy.authorize(admin_context, uppercase_action, self.target)

    @ddt.data('enforce', 'authorize')
    def test_authorize_properly_handles_invalid_scope_exception(self, method):
        self.fixture.config(enforce_scope=True, group='oslo_policy')
        project_context = context.RequestContext(project_id='fake-project-id',
                                                 roles=['bar'])
        policy.reset()
        policy.init(suppress_deprecation_warnings=True)
        rule = common_policy.RuleDefault('foo', 'role:bar',
                                         scope_types=['system'])
        policy._ENFORCER.register_defaults([rule])

        self.assertRaises(exception.PolicyNotAuthorized,
                          getattr(policy, method),
                          project_context, 'foo', {})

    @ddt.data('enforce', 'authorize')
    def test_authorize_does_not_raise_forbidden(self, method):
        self.fixture.config(enforce_scope=False, group='oslo_policy')
        project_context = context.RequestContext(project_id='fake-project-id',
                                                 roles=['bar'])
        policy.reset()
        policy.init(suppress_deprecation_warnings=True)
        rule = common_policy.RuleDefault('foo', 'role:bar',
                                         scope_types=['system'])
        policy._ENFORCER.register_defaults([rule])

        self.assertTrue(getattr(policy, method)(project_context, 'foo', {}))


class DefaultPolicyTestCase(test.TestCase):
    """This test case calls into the "enforce" method in policy

    enforce() in contrast with authorize() allows "default" rules to apply
    to policies that have not been registered.
    """

    def setUp(self):
        super(DefaultPolicyTestCase, self).setUp()
        policy.reset()
        policy.init(suppress_deprecation_warnings=True)

        self.rules = {
            "default": [],
            "example:exist": "false:false"
        }
        self._set_rules('default')
        self.context = context.RequestContext('fake', 'fake')

    def tearDown(self):
        super(DefaultPolicyTestCase, self).tearDown()
        policy.reset()

    def _set_rules(self, default_rule):
        these_rules = common_policy.Rules.from_dict(self.rules,
                                                    default_rule=default_rule)
        policy._ENFORCER.set_rules(these_rules)

    def test_policy_called(self):
        self.assertRaises(exception.PolicyNotAuthorized, policy.enforce,
                          self.context, "example:exist", {})

    def test_not_found_policy_calls_default(self):
        policy.enforce(self.context, "example:noexist", {})

    def test_default_not_found(self):
        new_default_rule = "default_noexist"
        # FIXME(gyee): need to overwrite the Enforcer's default_rule first
        # as it is recreating the rules with its own default_rule instead
        # of the default_rule passed in from set_rules(). I think this is a
        # bug in Oslo policy.
        policy._ENFORCER.default_rule = new_default_rule
        self._set_rules(new_default_rule)
        self.assertRaises(exception.PolicyNotAuthorized, policy.enforce,
                          self.context, "example:noexist", {})


class ContextIsAdminPolicyTestCase(test.TestCase):

    def setUp(self):
        super(ContextIsAdminPolicyTestCase, self).setUp()
        policy.reset()
        policy.init(suppress_deprecation_warnings=True)

    def _set_rules(self, rules, default_rule):
        these_rules = common_policy.Rules.from_dict(rules,
                                                    default_rule=default_rule)
        policy._ENFORCER.set_rules(these_rules)

    def test_default_admin_role_is_admin(self):
        ctx = context.RequestContext('fake', 'fake', roles=['johnny-admin'])
        self.assertFalse(ctx.is_admin)
        ctx = context.RequestContext('fake', 'fake', roles=['admin'])
        self.assertTrue(ctx.is_admin)

    def test_custom_admin_role_is_admin(self):
        # define explicit rules for context_is_admin
        rules = {
            'context_is_admin': [["role:administrator"], ["role:johnny-admin"]]
        }
        self._set_rules(rules, CONF.oslo_policy.policy_default_rule)
        ctx = context.RequestContext('fake', 'fake', roles=['johnny-admin'])
        self.assertTrue(ctx.is_admin)
        ctx = context.RequestContext('fake', 'fake', roles=['administrator'])
        self.assertTrue(ctx.is_admin)
        # default rule no longer applies
        ctx = context.RequestContext('fake', 'fake', roles=['admin'])
        self.assertFalse(ctx.is_admin)

    def test_context_is_admin_undefined(self):
        rules = {
            "admin_or_owner": "role:admin or project_id:%(project_id)s",
            "default": "rule:admin_or_owner",
        }
        self._set_rules(rules, CONF.oslo_policy.policy_default_rule)
        ctx = context.RequestContext('fake', 'fake')
        self.assertTrue(ctx.is_admin)
        ctx = context.RequestContext('fake', 'fake', roles=['admin'])
        self.assertTrue(ctx.is_admin)
